import datetime
import json, logging, asyncio, grpc, time, sys
from typing import Optional

from .grpc_stubs import device_agent_pb2, device_agent_pb2_grpc
from ..grpc_interface import GRPCInterface
from ...utils import maybe_async, call_maybe_async
from ...cli.decorators import command as cli_command, annotate_arg


class DeviceAgentInterface(GRPCInterface):
    stub = device_agent_pb2_grpc.deviceAgentStub

    def __init__(
        self,
        dda_uri: str = "127.0.0.1:50051",
        is_async: bool = None,
        dda_timeout: int = 7,
        max_conn_attempts: int = 5,
        time_between_connection_attempts: int = 10,
    ):
        super().__init__(dda_uri, is_async, dda_timeout)
        
        self.dda_timeout = dda_timeout
        self.max_connection_attempts = max_conn_attempts
        self.time_between_connection_attempts = time_between_connection_attempts

        self.is_dda_available = False
        self.is_dda_online = False
        self.has_dda_been_online = False

        self.subscriptions = {}
        # this is a list of channels that the agent interface will subscribe to,
        # and a list of callbacks that will be called when a message is received,
        # as well as the aggregate data that is received from the channel
        ## for channel in default_subscriptions:
            # #self.subscriptions[channel] = {"callbacks": [], "aggregate": None, is_synced: False}

        self.last_channel_message_ts = {}
        # this is a dictionary of the last time a message was received from a channel

        self.subscription_handlers = []
        # this is a list of async functions that will be called when the agent interface starts a subscription

    ## A helper to determine if the connection is persistent
    def has_persistent_connection(self):
        return True

    @cli_command()
    def get_is_dda_available(self):
        return self.is_dda_available
    
    @cli_command()
    def get_is_dda_online(self):
        return self.is_dda_online
    
    @cli_command()
    def get_has_dda_been_online(self):
        return self.has_dda_been_online

    def add_subscription(self, channel_name, callback=None):
        existing_channel_subs = self.subscriptions.keys()
        if not channel_name in existing_channel_subs:
            logging.debug("Adding subscription to channel: " + channel_name)
            listener = asyncio.ensure_future(self.start_subscription_listener(channel_name))
            self.subscription_handlers += [listener]
            self.subscriptions[channel_name] = {"listener": listener, "callbacks": [], "aggregate": None, "is_synced": False}
        else:
            logging.debug("Subscription already exists for channel: " + channel_name)

        if callback is not None:
            logging.debug("Adding new callback for subscription to " + channel_name)
            self.subscriptions[channel_name]["callbacks"].append(callback)

    async def start_subscription_listener(self, channel_name):
        while True:
            try:
                logging.debug("Starting subscription to channel: " + channel_name)
                await self.recv_channel_msgs(channel_name)
            except Exception as e:
                logging.error("Error starting subscription listener for " + str(channel_name) + ": " + str(e), exc_info=e)
                await asyncio.sleep(self.time_between_connection_attempts)

    async def recv_update_callback(self, channel_name, response):
        logging.debug("Received response from subscription request: " + str(response)[:100])
        resp = await self.make_request_async("GetChannelDetails", device_agent_pb2.ChannelDetailsRequest(channel_name=channel_name))
        if resp is None:
            logging.warning("Failed to get aggregate from channel: " + channel_name)
            return

        aggregate = resp.channel.aggregate
        success = resp.response_header.success

        ## validate aggregate is valid json
        try:
            json_aggregate = json.loads(aggregate)
            aggregate = json_aggregate
            logging.debug("Parsed aggregate from channel: " + channel_name + " : " + str(json_aggregate)[:100])
        except:
            logging.debug("Failed to parse aggregate from channel: " + channel_name)

        logging.debug("Received aggregate from channel: " + channel_name)
        self.subscriptions[channel_name]["aggregate"] = aggregate
        self.subscriptions[channel_name]["is_synced"] = success and aggregate not in [None, 'None']
        self.last_channel_message_ts[channel_name] = datetime.datetime.now()

        ## invoke all callbacks
        tasks = []
        for callback in self.subscriptions[channel_name]["callbacks"]:
            task = await call_maybe_async(callback, channel_name, aggregate, as_task=True)
            tasks.append(task)
        await asyncio.gather(*tasks)

    async def recv_channel_msgs(self, channel_name):

        ## Setup the connection to the doover device agent (DDA)
        async with grpc.aio.insecure_channel(self.uri) as channel:

            channel_stream = device_agent_pb2_grpc.deviceAgentStub(channel).GetChannelSubscription( device_agent_pb2.ChannelSubscriptionRequest(channel_name=channel_name))
            while True:
                try:
                    response = await channel_stream.read()
                    logging.debug("Received response from subscription request on " + str(channel_name) + " : " + str(response).replace("\n", " ")[:120])
                    self.update_dda_status(response.response_header)
                    if not response.response_header.success:
                        logging.error("Failed to subscribe to channel " + str(channel_name) + ": " + response.response_header.error_message)
                        return False
                    else:
                        logging.debug("Calling callback with subscription response for " + str(channel_name) + "...")
                        await self.recv_update_callback(channel_name, response)
                except StopAsyncIteration:
                    logging.debug("Channel stream ended.")
                    break

    def process_response(self, stub_call: str, response, *args, **kwargs):
        self.update_dda_status(response.response_header)
        return super().process_response(stub_call, response, *args, **kwargs)

    def update_dda_status(self, header):
        if header.success:
            self.is_dda_available = True
        else:
            self.is_dda_available = False

        if header.cloud_synced:
            self.is_dda_online = True
            if not self.has_dda_been_online:
                logging.info("Device Agent is online")
            self.has_dda_been_online = True
        else:
            self.is_dda_online = False

    def is_channel_synced(self, channel_name):
        if not channel_name in self.subscriptions:
            return False
        if not "is_synced" in self.subscriptions[channel_name]:
            return False
        return self.subscriptions[channel_name]["is_synced"]
    
    def wait_for_channels_sync(self, channel_names, timeout=5, inter_wait=0.2):
        start_time = datetime.datetime.now()
        while not all([self.is_channel_synced(channel_name) for channel_name in channel_names]):
            if (datetime.datetime.now() - start_time).seconds > timeout:
                return False
            time.sleep(inter_wait)
        return True

    async def wait_for_channels_sync_async(self, channel_names, timeout=5, inter_wait=0.2):
        start_time = datetime.datetime.now()
        while not all([self.is_channel_synced(channel_name) for channel_name in channel_names]):
            if (datetime.datetime.now() - start_time).seconds > timeout:
                return False
            await asyncio.sleep(inter_wait)
        return True

    @cli_command()
    @maybe_async()
    def get_channel_aggregate(self, channel_name):
        """Publish a message to a channel.

        Parameters
        ----------
        channel_name : str
            Name of channel to get aggregate from.
            
        Returns
        -------
        dict | str
            Aggregate data from channel
        """
        resp = self.make_request("GetChannelDetails", device_agent_pb2.ChannelDetailsRequest(channel_name=channel_name))
        return resp and resp.channel.aggregate

    async def get_channel_aggregate_async(self, channel_name):
        resp = await self.make_request_async("GetChannelDetails", device_agent_pb2.ChannelDetailsRequest(channel_name=channel_name))
        return resp and resp.channel.aggregate

    @cli_command()
    @maybe_async()
    def publish_to_channel(self, channel_name:str, message: dict | str, record_log: bool = True, save_aggregate: bool = False):
        """Publish a message to a channel.

        Parameters
        ----------
        channel_name : str
            Name of channel to publish data too.
        message : dict or str
            The data to send either in a dictionary or string format
        record_log : 
            Whether to save to the log
            
        Returns
        -------
        bool
            Whether publishing was successfull
        """
        if isinstance(message, dict):
            message = json.dumps(message)

        req = device_agent_pb2.ChannelWriteRequest(channel_name=channel_name, message_payload=message, save_log=record_log)
        resp = self.make_request("WriteToChannel", req)
        return resp and resp.response_header.success or False

    async def publish_to_channel_async(self, channel_name, message, record_log=True, save_aggregate=False):
        if isinstance(message, dict):
            message = json.dumps(message)

        req = device_agent_pb2.ChannelWriteRequest(channel_name=channel_name, message_payload=message, save_log=record_log)
        resp = await self.make_request_async("WriteToChannel", req)
        return resp and resp.response_header.success or False

    @staticmethod
    def _parse_get_token_response(response) -> Optional[tuple[str, datetime.datetime, str]]:
        if response is None:
            return None

        try:
            return response.token, datetime.datetime.fromtimestamp(float(response.valid_until)), response.endpoint
        except (ValueError, Exception) as e:
            logging.error("Failed to parse output from get_temp_token", exc_info=e)
            return None
    
    @cli_command()
    @maybe_async()
    def get_temp_token(self) -> Optional[tuple[str, datetime.datetime, str]]:
        """Get a temporary api token.

        Returns
        -------
        tuple
            (token, expire_time, url_endpoint)
        """
        resp = self.make_request("GetTempAPIToken", device_agent_pb2.TempAPITokenRequest())
        return self._parse_get_token_response(resp)

    async def get_temp_token_async(self) -> Optional[tuple[str, datetime.datetime, str]]:
        resp = await self.make_request_async("GetTempAPIToken", device_agent_pb2.TempAPITokenRequest())
        return self._parse_get_token_response(resp)

    def close(self):
        for listener in self.subscription_handlers:
            listener.cancel()
        logging.info("Closing device agent interface...")
        
    @cli_command()
    def test_comms(self, message: str = "Comms Check Message") -> Optional[str]:
        """Test connection by sending a basic echo response to device agent container.

        Parameters
        ----------
        message : str
            Message to send to device agent to have echo'd as a response

        Returns
        -------
        str
            The response from device agent.
        """
        return self.make_request("TestComms", device_agent_pb2.TestCommsRequest(message=message), response_field="response")

    @cli_command()
    def listen_channel(self, channel_name: str) -> Optional[str]:
        """Listen to channel printing the output to the console.

        Parameters
        ----------
        channel_name : str
            Name of channel to get aggregate from.

        Returns
        -------
        None
            Response is printed to stdout directly
        """
                
        loop = asyncio.get_event_loop()

        # If the loop is already running, create a background task
        if loop.is_running():
            asyncio.create_task(self.run_channel_listening(channel_name))
        else:
            try:
                loop.run_until_complete(self.run_channel_listening(channel_name))
            except KeyboardInterrupt:
                print("\nKeyboard interrupt detected. Exiting gracefully...")
            finally:
                loop.close()

    async def run_channel_listening(self, channel_name: str):
        
        def callback(channel_name, aggregate):
            print(channel_name, json.dumps(aggregate))
            sys.stdout.flush()
        
        self.add_subscription(channel_name, callback)
        
        try:
            while True:
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            self.close()

device_agent_iface = DeviceAgentInterface


from .channel import Channel


class Agent:

    def __init__(self, client, data):
        self.client = client

        self.id = None
        self.type = None
        self.name = None
        self.owner_org = None
        self.deployment_config = None
        self.channels = None

        self._from_data(data)

    def _from_data(self, data):
        # {'agent': '9fb5d629-ce7f-4b08-b17a-c267cbcd0427', 'current_time': 1715662485.485132, 'type': 'doover_users | user', 'channels': [{'channel': 'd1c7e8e3-f47b-4c68-86d7-65054d9e97d3', 'name': 'josh-test', 'type': 'base', 'agent': '9fb5d629-ce7f-4b08-b17a-c267cbcd0427'}, {'channel': '1f71b8bd-9444-4f34-859f-f339875a765c', 'name': 'test-logs', 'type': 'base', 'agent': '9fb5d629-ce7f-4b08-b17a-c267cbcd0427'}, {'channel': '86c96181-425d-46f4-a4eb-b36bde2b3984', 'name': 'notifications', 'type': 'base', 'agent': '9fb5d629-ce7f-4b08-b17a-c267cbcd0427'}]}
        self.id = data["agent"]
        self.type = data["type"]

        if 'name' in data: self.name = data["name"] 
        else: self.name = None

        if 'owner_org' in data: self.owner_org = data["owner_org"]
        else: self.owner_org = None

        if 'settings' in data and 'deployment_config' in data['settings']: self.deployment_config = data["settings"]["deployment_config"]
        else: self.deployment_config = None

        self.channels = [Channel(data=c, client=self.client) for c in data["channels"]]

    @property
    def agent_id(self):
        return self.id

    def update(self):
        res = self.client._get_agent_raw(self.id)
        return self._from_data(res)



#!/usr/bin/env python3

import asyncio
import logging

from typing import Optional

import grpc

from .grpc_stubs import modbus_iface_pb2, modbus_iface_pb2_grpc
from ..grpc_interface import GRPCInterface
from ...utils import call_maybe_async, maybe_async
from ...cli.decorators import command as cli_command, ignore_alias

# just lower case should be fine...
VALID_SERIAL_CONFIG_KEYS = (
    "serial_port", "serial_baud", "serial_method", "serial_bits", "serial_parity", "serial_stop", "serial_timeout",
    "SERIAL_PORT", "SERIAL_BAUD", "SERIAL_METHOD", "SERIAL_BITS", "SERIAL_PARITY", "SERIAL_STOP", "SERIAL_TIMEOUT",
)


def two_words_to_32bit_float(word1, word2, swap=False):
    if swap:
        word1, word2 = word2, word1
    return word1 + (word2 * 65536)


class ModbusInterface(GRPCInterface):
    stub = modbus_iface_pb2_grpc.modbusIfaceStub

    def __init__(self, modbus_uri: str = "127.0.0.1:50054", is_async: bool = None, config_manager=None):
        super().__init__(modbus_uri, is_async)

        self.subscription_tasks = []

        self.setup_task = None
        self.config_complete = True
        if config_manager is not None:
            self.set_config_manager(config_manager=config_manager)
            self.setup_task = asyncio.create_task(self.setup_from_config_manager())


    def set_config_manager(self, config_manager, run_setup=False):
        if self.setup_task is not None:
            logging.warning("Modbus Iface has already been setup, skipping new config manager setup")
            # fixme: should this return None here?

        self.config_complete = False
        self.config_manager = config_manager

        if run_setup:
            self.setup_task = asyncio.create_task(self.setup_from_config_manager())

    async def setup_from_config_manager(self):
        await self.config_manager.await_config()

        # explicitly call async varient because we can
        modbus_config = await self.config_manager.get_config_async('MODBUS_CONFIG')
        logging.info("Setting up modbus iface with following config : " + str(modbus_config))
        if modbus_config is None:
            logging.info("No modbus config found")
            return None
        
        if "SERIAL_PORT" in modbus_config:
            config = {k.lower(): v for k, v in modbus_config.items() if k in VALID_SERIAL_CONFIG_KEYS}
            logging.info("opening new modbus bus on serial port with configuration %s", str(config))
            await self.open_bus_async(bus_type="serial", name="default", **config)

        elif "TCP_URI" in modbus_config:
            tcp_uri = modbus_config.get("TCP_URI", None)
            tcp_timeout = modbus_config.get("TCP_TIMEOUT", None)

            logging.info("opening new tcp modbus bus on uri " + str(tcp_uri))

            await self.open_bus_async(
                bus_type="tcp",
                name="default",
                tcp_uri=tcp_uri,
                tcp_timeout=tcp_timeout,
            )

        elif "BUSES" in modbus_config:

            logging.info("opening multiple modbus buses from config")

            for bus in modbus_config["BUSES"]:
                bus_type = bus.get("BUS_TYPE", None)
                bus_name = bus.get("BUS_NAME", None)

                if bus_type == "serial":
                    serial_port = bus.get("SERIAL_PORT", None)
                    serial_baud = bus.get("SERIAL_BAUD", None)
                    serial_method = bus.get("SERIAL_METHOD", None)
                    serial_bits = bus.get("SERIAL_BITS", None)
                    serial_parity = bus.get("SERIAL_PARITY", None)
                    serial_stop = bus.get("SERIAL_STOP", None)
                    serial_timeout = bus.get("SERIAL_TIMEOUT", None)

                    logging.info("opening new modbus bus on serial port " + str(serial_port))

                    await self.open_bus_async(
                        bus_type="serial",
                        name=bus_name,
                        serial_port=serial_port,
                        serial_baud=serial_baud,
                        serial_method=serial_method,
                        serial_bits=serial_bits,
                        serial_parity=serial_parity,
                        serial_stop=serial_stop,
                        serial_timeout=serial_timeout,
                    )

                elif bus_type == "tcp":
                    tcp_uri = bus.get("TCP_URI", None)
                    tcp_timeout = bus.get("TCP_TIMEOUT")

                    logging.info("opening new tcp modbus bus on uri " + str(tcp_uri))

                    await self.open_bus_async(
                        bus_type="tcp",
                        name=bus_name,
                        tcp_uri=tcp_uri,
                        tcp_timeout=tcp_timeout,
                    )
        
        else:
            logging.info("No modbus buses opened from config")

        self.config_complete = True

    def process_response(self, stub_call: str, response, *args, **kwargs):
        resp = super().process_response(stub_call, response, *args, **kwargs)

        # only for some of the calls we want to ensure a bus is available (e.g. read/write registers)
        # this will fix it for next run but not the current one...
        try:
            configure_bus = kwargs["configure_bus"]
            bus_id = kwargs["bus_id"]

            if not response.response_header.success and configure_bus and not self.ensure_bus_availabe(bus_id, response.response_header):
                logging.warning("Seems bus is not available")
        except KeyError:
            pass

        return resp

    def ensure_bus_availabe(self, bus_id, response_header, configure=True):
        ## if not config_complete, wait for setup to complete
        if not self.config_complete or (self.setup_task is not None and not self.setup_task.done()):
            logging.debug("Waiting for modbus setup to complete")
            return False

        ## check the bus status from the response and if the bus does not exist, and configure is True, rerun the setup
        bus_statusses = response_header.bus_status
        for b in bus_statusses:
            if b.bus_id == bus_id:
                return b.open
            
        logging.warning("Bus " + str(bus_id) + " not found in response")
        if configure:
            logging.info("Reconfiguring modbus iface")
            asyncio.run(self.setup_from_config_manager())
        
        return True


    def close(self):
        logging.info("Closing modbus interface")
        for task in self.subscription_tasks:
            task.cancel()

    @staticmethod
    def _get_bus_request(
        bus_type="serial",
        name="default",
        serial_port="/dev/ttyS0",
        serial_baud=9600,
        serial_method="rtu",
        serial_bits=8,
        serial_parity="N",
        serial_stop=1,
        serial_timeout=0.3,
        tcp_uri="127.0.0.1:5000",
        tcp_timeout=2,
    ):
        if bus_type not in ("serial", "tcp"):
            logging.error("Invalid bus type: " + str(bus_type))
            return None

        kwargs = {"bus_id": str(name)}

        if bus_type == "serial":
            kwargs["serial_settings"] = modbus_iface_pb2.serialBusSettings(
                port=serial_port,
                baud=serial_baud,
                modbus_method=serial_method,
                data_bits=serial_bits,
                parity=serial_parity,
                stop_bits=serial_stop,
                timeout=serial_timeout,
            )
        elif bus_type == "tcp":
            ip, port = tcp_uri.split(":")
            kwargs["ethernet_settings"] = modbus_iface_pb2.ethernetBusSettings(
                ip=ip, port=int(port), timeout=tcp_timeout
            )
        else:
            logging.error("Invalid bus type: " + str(bus_type))
            return None

        return modbus_iface_pb2.openBusRequest(**kwargs)

    @cli_command()
    @maybe_async()
    def open_bus(
        self,
        bus_type="serial",
        name="default",
        serial_port="/dev/ttyS0",
        serial_baud=9600,
        serial_method="rtu",
        serial_bits=8,
        serial_parity="N",
        serial_stop=1,
        serial_timeout=0.3,
        tcp_uri="127.0.0.1:5000",
        tcp_timeout=2,
    ) -> bool:
        req = self._get_bus_request(
            bus_type, name, serial_port, serial_baud, serial_method, serial_bits,
            serial_parity, serial_stop, serial_timeout, tcp_uri, tcp_timeout,
        )
        if req is None:
            return

        resp = self.make_request("openBus", req)
        return resp.response_header.success

    async def open_bus_async(
        self,
        bus_type="serial",
        name="default",
        serial_port="/dev/ttyS0",
        serial_baud=9600,
        serial_method="rtu",
        serial_bits=8,
        serial_parity="N",
        serial_stop=1,
        serial_timeout=0.3,
        tcp_uri="127.0.0.1:5000",
        tcp_timeout=2,
    ) -> bool:
        req = self._get_bus_request(
            bus_type, name, serial_port, serial_baud, serial_method, serial_bits,
            serial_parity, serial_stop, serial_timeout, tcp_uri, tcp_timeout,
        )
        if req is None:
            return False

        resp = await self.make_request_async("openBus", req)
        return resp.response_header.success

    @cli_command()
    @maybe_async()
    def close_bus(self, bus_id: str = "default") -> bool:
        req = modbus_iface_pb2.closeBusRequest(bus_id=str(bus_id))
        resp = self.make_request("closeBus", req)
        return resp.response_header.success and resp.bus_status.open

    async def close_bus_async(self, bus_id: str = "default") -> bool:
        req = modbus_iface_pb2.closeBusRequest(bus_id=str(bus_id))
        resp = await self.make_request_async("closeBus", req)
        return resp.response_header.success and resp.bus_status.open

    def _validate_read_register_resp(self, resp, bus_id, configure_bus):
        try:
            if not resp.response_header.success:
                logging.error("Error reading registers from bus " + str(bus_id))
                return False
            # return self.ensure_bus_availabe(bus_id, resp.response_header, configure_bus)
            return True
        except Exception as e:
            logging.error("Error validating read register response: " + str(e))
            return False
    
    @cli_command()
    @maybe_async()
    def get_bus_status(self, bus_id: str = "default") -> bool:
        req = modbus_iface_pb2.busStatusRequest(bus_id=str(bus_id))
        resp = self.make_request("busStatus", req)
        return resp.response_header.success and resp.bus_status.open

    async def get_bus_status_async(self, bus_id: str = "default") -> bool:
        req = modbus_iface_pb2.busStatusRequest(bus_id=str(bus_id))
        resp = await self.make_request_async("busStatus", req)
        return resp.response_header.success and resp.bus_status.open

    getBusStatus = ignore_alias(get_bus_status)
    getBusStatus_async = get_bus_status_async

    @staticmethod
    def _parse_register_output(values):
        if len(values) == 0:
            return None
        if len(values) == 1:
            return values[0]
        return values

    @cli_command()
    @maybe_async()
    def read_registers(
        self,
        bus_id: str = "default",
        modbus_id: int = 1,
        start_address: int = 0,
        num_registers: int = 1,
        register_type: int = 4,
        configure_bus: bool = True,
    ) -> Optional[int | list[int]]:
        req = modbus_iface_pb2.readRegisterRequest(
            bus_id=str(bus_id),
            modbus_id=modbus_id,
            register_type=register_type,
            address=start_address,
            count=num_registers,
        )
        resp = self.make_request("readRegisters", req, bus_id=bus_id, configure_bus=configure_bus)
        return resp and self._parse_register_output(resp.values)

    async def read_registers_async(
        self,
        bus_id: str = "default",
        modbus_id: int = 1,
        start_address: int = 0,
        num_registers: int = 1,
        register_type: int = 4,
        configure_bus: bool = True,
    ) -> Optional[int | list[int]]:
        req = modbus_iface_pb2.readRegisterRequest(
            bus_id=str(bus_id),
            modbus_id=modbus_id,
            register_type=register_type,
            address=start_address,
            count=num_registers,
        )
        resp = await self.make_request_async("readRegisters", req, bus_id=bus_id, configure_bus=configure_bus)
        return resp and self._parse_register_output(resp.values)

    @cli_command()
    @maybe_async()
    def write_registers(
        self,
        bus_id: str = "default",
        modbus_id: int = 1,
        start_address: int = 0,
        values: list[int] = None,
        register_type: int = 4,
        configure_bus=True,
    ) -> bool:
        values = values or []
        req = modbus_iface_pb2.writeRegisterRequest(
            bus_id=str(bus_id),
            modbus_id=modbus_id,
            register_type=register_type,
            address=start_address,
            values=values,
        )
        resp = self.make_request("writeRegisters", req, bus_id=bus_id, configure_bus=configure_bus)
        return resp and self._validate_read_register_resp(resp, bus_id, configure_bus)

    async def write_registers_async(
        self,
        bus_id: str = "default",
        modbus_id: int = 1,
        start_address: int = 0,
        values: list[int] = None,
        register_type: int = 4,
        configure_bus=True,
    ) -> bool:
        values = values or []
        req = modbus_iface_pb2.writeRegisterRequest(
            bus_id=str(bus_id),
            modbus_id=modbus_id,
            register_type=register_type,
            address=start_address,
            values=values,
        )
        resp = await self.make_request_async("writeRegisters", req, bus_id=bus_id, configure_bus=configure_bus)
        return resp and self._validate_read_register_resp(resp, bus_id, configure_bus)


    def add_read_register_subscription(
            self, 
            bus_id="default",
            modbus_id=1,
            start_address=0,
            num_registers=1,
            register_type=4,
            poll_secs=3,
            callback=None,
        ):

        if callback is None:
            logging.error("No callback provided for read register subscription")
            return None
        
        try:
            new_task = asyncio.create_task(self.run_read_register_subscription_task(
                bus_id=str(bus_id),
                modbus_id=modbus_id,
                start_address=start_address,
                num_registers=num_registers,
                register_type=register_type,
                poll_secs=poll_secs,
                callback=callback,
            ))

            self.subscription_tasks.append(new_task)
            new_task.add_done_callback(self.subscription_tasks.remove)
            return new_task
        
        except Exception as e:
            logging.error("Error adding read register subscription: " + str(e))
            return None


    async def run_read_register_subscription_task(
            self,
            bus_id="default",
            modbus_id=1,
            start_address=0,
            num_registers=1,
            register_type=4,
            poll_secs=3,
            callback=None,
            configure_bus=True,
        ):

        try:
            async with grpc.aio.insecure_channel(self.uri) as channel:
                stub = modbus_iface_pb2_grpc.modbusIfaceStub(channel)
                request = modbus_iface_pb2.readRegisterSubscriptionRequest(
                    bus_id=str(bus_id),
                    modbus_id=modbus_id,
                    register_type=register_type,
                    address=start_address,
                    count=num_registers,
                    poll_secs=poll_secs,
                )

                try:
                    async for response in stub.readRegisterSubscription(request):
                        
                        success = response.response_header.success
                        if not self._validate_read_register_resp(response, bus_id, configure_bus):
                            values = None
                        elif len(response.values) == 1:
                            values = response.values[0]
                        else:
                            values = response.values

                        logging.debug("recieved new modbus subscription result on bus " + str(bus_id) + ", for modbus_id " + str(modbus_id) + ", result=" + str(success))
                        if callback is not None:
                            await call_maybe_async(callback, values)

                except Exception as e:
                    logging.error("Error in read register subscription task: " + str(e))
                    return None
        
        except Exception as e:
            logging.error("Error in read register subscription task: " + str(e))
            return None
        
    @cli_command()
    def test_comms(self, message: str = "Comms Check Message") -> Optional[str]:
        """Test connection by sending a basic echo response to modbus interface container.

        Parameters
        ----------
        message : str
            Message to send to modbus interface to have echo'd as a response

        Returns
        -------
        str
            The response from modbus interface.
        """
        return self.make_request("testComms", modbus_iface_pb2.testCommsRequest(message=message), response_field="response")



modbus_iface = ModbusInterface
# async def run_test():
    
if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG)

    iface = modbus_iface()
    iface.open_bus(
        bus_type="serial",
        name="test",
        serial_port="/dev/ttyk37simOut",
        serial_baud=38400,
        serial_method="rtu",
        serial_bits=8,
        serial_parity="N",
        serial_stop=1,
        serial_timeout=0.3,
    )

    print(iface.getBusStatus(bus_id="test"))

    result = iface.read_registers(
        bus_id="test",
        modbus_id=1,
        start_address=0,
        num_registers=23,
    )
    print(result)

    watchdog = result[22] + 1

    result = iface.write_registers(
        bus_id="test",
        modbus_id=1,
        start_address=22,
        values=[watchdog],
    )

    ## define a function to print the results of read register subscription
    def print_results(values):
        print(values)

    loop = asyncio.get_event_loop()

    ## add a read register subscription
    iface.add_read_register_subscription(
        bus_id="test",
        modbus_id=1,
        start_address=0,
        num_registers=23,
        callback=print_results,
    )

    print("Subscribed to read register subscription")

    ## add a read register subscription
    iface.add_read_register_subscription(
        bus_id="test",
        modbus_id=1,
        start_address=0,
        num_registers=10,
        callback=print_results,
    )

    # async def run_test():
    #     await asyncio.sleep(20)
    #     iface.close()

    try:
        asyncio.get_event_loop().run_forever()
    except KeyboardInterrupt:
        pass

    logging.info("Closing modbus interface")
    iface.close()





# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import modbus_iface_pb2 as grpc__stubs_dot_modbus__iface__pb2

GRPC_GENERATED_VERSION = '1.65.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.66.0'
SCHEDULED_RELEASE_DATE = 'August 6, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in grpc_stubs/modbus_iface_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class modbusIfaceStub(object):
    """The doover modbus service definition.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.testComms = channel.unary_unary(
                '/modbus_iface.modbusIface/testComms',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.testCommsRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.testCommsResponse.FromString,
                _registered_method=True)
        self.openBus = channel.unary_unary(
                '/modbus_iface.modbusIface/openBus',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.openBusRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.openBusResponse.FromString,
                _registered_method=True)
        self.listBus = channel.unary_unary(
                '/modbus_iface.modbusIface/listBus',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.listBusRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.listBusResponse.FromString,
                _registered_method=True)
        self.busStatus = channel.unary_unary(
                '/modbus_iface.modbusIface/busStatus',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.busStatusRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.busStatusResponse.FromString,
                _registered_method=True)
        self.closeBus = channel.unary_unary(
                '/modbus_iface.modbusIface/closeBus',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.closeBusRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.closeBusResponse.FromString,
                _registered_method=True)
        self.readRegisters = channel.unary_unary(
                '/modbus_iface.modbusIface/readRegisters',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.readRegisterRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.readRegisterResponse.FromString,
                _registered_method=True)
        self.writeRegisters = channel.unary_unary(
                '/modbus_iface.modbusIface/writeRegisters',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.writeRegisterRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.writeRegisterResponse.FromString,
                _registered_method=True)
        self.scheduleWriteRegisters = channel.unary_unary(
                '/modbus_iface.modbusIface/scheduleWriteRegisters',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.scheduleWriteRegisterRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.scheduleWriteRegisterResponse.FromString,
                _registered_method=True)
        self.readRegisterSubscription = channel.unary_stream(
                '/modbus_iface.modbusIface/readRegisterSubscription',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.readRegisterSubscriptionRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.readRegisterSubscriptionResponse.FromString,
                _registered_method=True)
        self.createServer = channel.unary_unary(
                '/modbus_iface.modbusIface/createServer',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.createServerRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.createServerResponse.FromString,
                _registered_method=True)
        self.listServer = channel.unary_unary(
                '/modbus_iface.modbusIface/listServer',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.listServerRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.listServerResponse.FromString,
                _registered_method=True)
        self.closeServer = channel.unary_unary(
                '/modbus_iface.modbusIface/closeServer',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.closeServerRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.closeServerResponse.FromString,
                _registered_method=True)
        self.setServerRegisters = channel.unary_unary(
                '/modbus_iface.modbusIface/setServerRegisters',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.setServerRegistersRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.setServerRegistersResponse.FromString,
                _registered_method=True)
        self.getServerRegisters = channel.unary_unary(
                '/modbus_iface.modbusIface/getServerRegisters',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.getServerRegistersRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.getServerRegistersResponse.FromString,
                _registered_method=True)
        self.scheduleServerRegisters = channel.unary_unary(
                '/modbus_iface.modbusIface/scheduleServerRegisters',
                request_serializer=grpc__stubs_dot_modbus__iface__pb2.scheduleServerRegistersRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_modbus__iface__pb2.scheduleServerRegistersResponse.FromString,
                _registered_method=True)


class modbusIfaceServicer(object):
    """The doover modbus service definition.
    """

    def testComms(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def openBus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listBus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def busStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def closeBus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def readRegisters(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def writeRegisters(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def scheduleWriteRegisters(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def readRegisterSubscription(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def createServer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listServer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def closeServer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setServerRegisters(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getServerRegisters(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def scheduleServerRegisters(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_modbusIfaceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'testComms': grpc.unary_unary_rpc_method_handler(
                    servicer.testComms,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.testCommsRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.testCommsResponse.SerializeToString,
            ),
            'openBus': grpc.unary_unary_rpc_method_handler(
                    servicer.openBus,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.openBusRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.openBusResponse.SerializeToString,
            ),
            'listBus': grpc.unary_unary_rpc_method_handler(
                    servicer.listBus,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.listBusRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.listBusResponse.SerializeToString,
            ),
            'busStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.busStatus,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.busStatusRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.busStatusResponse.SerializeToString,
            ),
            'closeBus': grpc.unary_unary_rpc_method_handler(
                    servicer.closeBus,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.closeBusRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.closeBusResponse.SerializeToString,
            ),
            'readRegisters': grpc.unary_unary_rpc_method_handler(
                    servicer.readRegisters,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.readRegisterRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.readRegisterResponse.SerializeToString,
            ),
            'writeRegisters': grpc.unary_unary_rpc_method_handler(
                    servicer.writeRegisters,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.writeRegisterRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.writeRegisterResponse.SerializeToString,
            ),
            'scheduleWriteRegisters': grpc.unary_unary_rpc_method_handler(
                    servicer.scheduleWriteRegisters,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.scheduleWriteRegisterRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.scheduleWriteRegisterResponse.SerializeToString,
            ),
            'readRegisterSubscription': grpc.unary_stream_rpc_method_handler(
                    servicer.readRegisterSubscription,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.readRegisterSubscriptionRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.readRegisterSubscriptionResponse.SerializeToString,
            ),
            'createServer': grpc.unary_unary_rpc_method_handler(
                    servicer.createServer,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.createServerRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.createServerResponse.SerializeToString,
            ),
            'listServer': grpc.unary_unary_rpc_method_handler(
                    servicer.listServer,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.listServerRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.listServerResponse.SerializeToString,
            ),
            'closeServer': grpc.unary_unary_rpc_method_handler(
                    servicer.closeServer,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.closeServerRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.closeServerResponse.SerializeToString,
            ),
            'setServerRegisters': grpc.unary_unary_rpc_method_handler(
                    servicer.setServerRegisters,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.setServerRegistersRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.setServerRegistersResponse.SerializeToString,
            ),
            'getServerRegisters': grpc.unary_unary_rpc_method_handler(
                    servicer.getServerRegisters,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.getServerRegistersRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.getServerRegistersResponse.SerializeToString,
            ),
            'scheduleServerRegisters': grpc.unary_unary_rpc_method_handler(
                    servicer.scheduleServerRegisters,
                    request_deserializer=grpc__stubs_dot_modbus__iface__pb2.scheduleServerRegistersRequest.FromString,
                    response_serializer=grpc__stubs_dot_modbus__iface__pb2.scheduleServerRegistersResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'modbus_iface.modbusIface', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('modbus_iface.modbusIface', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class modbusIface(object):
    """The doover modbus service definition.
    """

    @staticmethod
    def testComms(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/testComms',
            grpc__stubs_dot_modbus__iface__pb2.testCommsRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.testCommsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def openBus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/openBus',
            grpc__stubs_dot_modbus__iface__pb2.openBusRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.openBusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def listBus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/listBus',
            grpc__stubs_dot_modbus__iface__pb2.listBusRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.listBusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def busStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/busStatus',
            grpc__stubs_dot_modbus__iface__pb2.busStatusRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.busStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def closeBus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/closeBus',
            grpc__stubs_dot_modbus__iface__pb2.closeBusRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.closeBusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def readRegisters(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/readRegisters',
            grpc__stubs_dot_modbus__iface__pb2.readRegisterRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.readRegisterResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def writeRegisters(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/writeRegisters',
            grpc__stubs_dot_modbus__iface__pb2.writeRegisterRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.writeRegisterResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def scheduleWriteRegisters(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/scheduleWriteRegisters',
            grpc__stubs_dot_modbus__iface__pb2.scheduleWriteRegisterRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.scheduleWriteRegisterResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def readRegisterSubscription(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/modbus_iface.modbusIface/readRegisterSubscription',
            grpc__stubs_dot_modbus__iface__pb2.readRegisterSubscriptionRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.readRegisterSubscriptionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def createServer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/createServer',
            grpc__stubs_dot_modbus__iface__pb2.createServerRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.createServerResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def listServer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/listServer',
            grpc__stubs_dot_modbus__iface__pb2.listServerRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.listServerResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def closeServer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/closeServer',
            grpc__stubs_dot_modbus__iface__pb2.closeServerRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.closeServerResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def setServerRegisters(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/setServerRegisters',
            grpc__stubs_dot_modbus__iface__pb2.setServerRegistersRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.setServerRegistersResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getServerRegisters(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/getServerRegisters',
            grpc__stubs_dot_modbus__iface__pb2.getServerRegistersRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.getServerRegistersResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def scheduleServerRegisters(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/modbus_iface.modbusIface/scheduleServerRegisters',
            grpc__stubs_dot_modbus__iface__pb2.scheduleServerRegistersRequest.SerializeToString,
            grpc__stubs_dot_modbus__iface__pb2.scheduleServerRegistersResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import logging

from collections import namedtuple
from datetime import datetime, timedelta
from typing import Any, Union, Callable, overload, Literal, Optional, TypeVar
from urllib.parse import quote, urlencode

import requests

from .message import Message
from .agent import Agent
from .channel import Channel, Processor, Task
from .exceptions import NotFound, Forbidden, HTTPException


log = logging.getLogger(__name__)
AccessToken = namedtuple("AccessToken", ["token", "expires_at"], defaults=(None, ))
T = TypeVar("T", bound=Channel)


class Route:
    def __init__(self, method, route, *args, **kwargs):
        self.method = method

        self.url = route
        if args:
            self.url = route.format(*[quote(a) for a in args])

        if kwargs:
            self.url = f"{self.url}?{urlencode(kwargs)}"


class Client:
    """API Client for Doover Cloud"""

    def __init__(
        self,
        username: str = None,
        password: str = None,
        token: str = None,
        token_expires: datetime = None,
        base_url: str = "https://my.doover.dev",
        agent_id: str = None,
        verify: bool = True,
        login_callback: Callable = None,
        config_profile: str = "default",
        debug: bool = False,
    ):
        
        self.access_token = None
        if token:
            self.access_token = AccessToken(token, token_expires)
        
        self.agent_id = agent_id
        self.login_callback = login_callback

        self.username = username
        self.password = password

        self.verify = verify
        self.base_url = base_url
        self.session = requests.Session()

        self.request_retries = 1
        self.request_timeout = 59

        if debug:
            logging.basicConfig(level=logging.DEBUG)

        if not ((username and password) or token):
            # fixme: work out these circular imports properly...
            from ...cli.config import ConfigManager

            self.config_manager = ConfigManager(config_profile)
            self.config_manager.read()

            config = self.config_manager.current
            if not config:
                raise RuntimeError(f"No configuration found for profile {self.config_manager.current_profile}. "
                                   f"Please specify a profile with the `config_profile` parameter, "
                                   f"manually set a token or `doover login`")

            self.agent_id = config.agent_id
            self.username = config.username
            self.password = config.password
            self.access_token = AccessToken(config.token, config.token_expires)
            self.base_url = config.base_url

        if self.access_token:
            self.update_headers()

    def has_persistent_connection(self):
        return False

    def update_headers(self):
        self.session.headers.update({"Authorization": f"Token {self.access_token.token}"})
        self.session.verify = self.verify

    def request(self, route: Route, **kwargs):
        # default is access token to not expire
        if self.access_token.expires_at and self.access_token.expires_at < datetime.utcnow():
            logging.info("Token expired, attempting to refresh token.")
            self.login()

        url = self.base_url + route.url

        attempt_counter = 0
        retries = self.request_retries if route.method == "GET" else 0

        while attempt_counter <= retries:
            attempt_counter += 1

            log.debug(f"Making {route.method} request to {url} with kwargs {kwargs}")
            
            try:
                resp = self.session.request(route.method, url, timeout=self.request_timeout, **kwargs)
            except requests.exceptions.Timeout:
                log.info(f"Request to {url} timed out.")
                if attempt_counter > retries:
                    raise HTTPException(f"Request timed out. {url}")
                continue

            if resp.status_code == 200:
                ## if we get a 200, we're good to go
                break
            elif resp.status_code == 403:
                raise Forbidden(f"Access denied. {url}")
            elif resp.status_code == 404:
                raise NotFound(f"Resource not found. {url}")
            elif resp.status_code != 200:
                log.info(f"Failed to make request to {url}. Status code: {resp.status_code}, message: {resp.text}")
                if attempt_counter > retries:
                    raise HTTPException(resp.text)

        try:
            data = resp.json()
        except ValueError:
            data = resp.text

        log.debug(f"{url} has received {data}")
        return data

    def _get_agent_raw(self, agent_id: str) -> dict[str, Any]:
        return self.request(Route("GET", "/ch/v1/agent/{}/", agent_id))

    def _get_agent_list_raw(self) -> list[dict[str, Any]]:
        return self.request(Route("GET", "/ch/v1/list_agents/"))

    def get_agent(self, agent_id: str) -> Optional[Agent]:
        data = self._get_agent_raw(agent_id)
        return data and Agent(client=self, data=data)

    def get_agent_list(self) -> list[Agent]:
        data = self._get_agent_list_raw()
        if not "agents" in data:
            return []
        return [Agent(client=self, data=d) for d in data["agents"]]

    def _parse_channel(self, data) -> T:
        if data["name"].startswith("!"):
            return Task(client=self, data=data)
        elif data["name"].startswith("#"):
            return Processor(client=self, data=data)
        else:
            return Channel(client=self, data=data)

    def _get_channel_raw(self, channel_id: str) -> dict[str, Any]:
        return self.request(Route("GET", "/ch/v1/channel/{}/", channel_id))

    def get_channel(self, channel_id: str) -> Optional[T]:
        data = self._get_channel_raw(channel_id)
        return data and self._parse_channel(data)

    def _get_channel_named_raw(self, channel_name: str, agent_id: str) -> dict[str, Any]:
        return self.request(Route("GET", "/ch/v1/agent/{}/{}/", agent_id, channel_name))

    def get_channel_named(self, channel_name: str, agent_id: str) -> Optional[T]:
        data = self._get_channel_named_raw(channel_name, agent_id)
        return data and self._parse_channel(data)

    def get_channel_messages(self, channel_id: str, num_messages: Optional[int] = None) -> list[Message]:
        if num_messages:
            data = self.request(Route("GET", "/ch/v1/channel/{}/messages/{}/", channel_id, str(num_messages)))
        else:
            data = self.request(Route("GET", "/ch/v1/channel/{}/messages/", channel_id))

        if not data:
            return []

        return [Message(client=self, data=m, channel_id=channel_id) for m in data["messages"]]

    def _get_message_raw(self, channel_id: str, message_id: str) -> dict[str, Any]:
        return self.request(Route("GET", "/ch/v1/channel/{}/message/{}", channel_id, message_id))

    def get_channel_messages_in_window(self, channel_id: str, start: datetime, end: datetime) -> list[Message]:
        start = str(int(start.timestamp()))
        end = str(int(end.timestamp()))
        data = self.request(Route("GET", "/ch/v1/channel/{}/messages/time/{}/{}/", channel_id, start, end))
        if not data:
            return []

        return [Message(client=self, data=m, channel_id=channel_id) for m in data["messages"]]

    def get_message(self, channel_id: str, message_id: str) -> Optional[Message]:
        data = self._get_message_raw(channel_id, message_id)
        return data and Message(client=self, data=data, channel_id=channel_id)

    def _delete_message_raw(self, channel_id: str, message_id: str) -> bool:
        return self.request(Route("DELETE", "/ch/v1/channel/{}/message/{}", channel_id, message_id))

    def create_channel(self, channel_name: str, agent_id: str) -> T:
        try:
            return self.get_channel_named(channel_name, agent_id)
        except NotFound:
            pass
        # all we need to do is publish to a channel with an empty payload
        self.request(Route("POST", "/ch/v1/agent/{}/{}/", agent_id, channel_name))
        # this is a bit of a wasted API call, but since this is the same method to post an aggregate to a
        # channel it can either return a new channel ID (if created), or the message ID of the posted message.
        return self.get_channel_named(channel_name, agent_id)

    def create_processor(self, processor_name: str, agent_id: str) -> Processor:
        return self.create_channel("#" + processor_name.lstrip('#'), agent_id)

    def create_task(self, task_name: str, agent_id: str, processor_id: str) -> Task:
        task = "!" + task_name.lstrip('!')
        payload = {
            "msg": {},  # this is a required field apparently
            "processor_id": processor_id
        }
        self.request(Route("POST", "/ch/v1/agent/{}/{}/", agent_id, task), json=payload)
        return self.get_channel_named(task, agent_id)

    def _maybe_subscribe_to_channel(self, channel_id: str, task_id: str, subscribe: bool):
        data = {"channel_id": channel_id, "subscribe": subscribe}
        return self.request(Route("POST", "/ch/v1/channel/{}/subscribe/", task_id), json=data)

    def subscribe_to_channel(self, channel_id: str, task_id: str) -> bool:
        return self._maybe_subscribe_to_channel(channel_id, task_id, True)

    def unsubscribe_from_channel(self, channel_id: str, task_id: str) -> bool:
        return self._maybe_subscribe_to_channel(channel_id, task_id, False)

    def publish_to_channel(self, channel_id: str, data: Any, save_log: bool = True, log_aggregate: bool = False, override_aggregate: bool = False, timestamp: Optional[datetime] = None):
        # basically we're assuming there's only 2 types of data - dict or string...
        post_data = {"msg": data}
        
        post_data["record_log"] = save_log
        if log_aggregate:
            post_data["log_aggregate"] = True
        if override_aggregate:
            post_data["override_aggregate"] = True
        if timestamp:
            post_data["timestamp"] = int(timestamp.timestamp())

        if isinstance(post_data, dict):
            return self.request(Route("POST", "/ch/v1/channel/{}/", channel_id), json=post_data)
        else:
            return self.request(Route("POST", "/ch/v1/channel/{}/", channel_id), data=str(post_data))

    def publish_to_channel_name(self, agent_id: str, channel_name: str, data: Any, save_log: bool = True, log_aggregate: bool = False, override_aggregate: bool = False, timestamp: Optional[datetime] = None):
        post_data = {"msg": data}
        
        post_data["record_log"] = save_log
        if log_aggregate:
            post_data["log_aggregate"] = True
        if override_aggregate:
            post_data["override_aggregate"] = True
        if timestamp:
            post_data["timestamp"] = int(timestamp.timestamp())
        
        if isinstance(post_data, dict):
            return self.request(Route("POST", "/ch/v1/agent/{}/{}/", agent_id, channel_name), json=post_data)
        else:
            return self.request(Route("POST", "/ch/v1/agent/{}/{}/", agent_id, channel_name), data=str(post_data))

    def create_tunnel_endpoints(self, agent_id: str, endpoint_type: str, amount: int):
        to_return = []
        for i in range(amount):
            res = self.request(Route("POST", "/ch/v1/agent/{}/ngrok_tunnels/{}/", agent_id, endpoint_type))
            if res and res.get("url"):
                to_return.append(res["url"])
        return to_return

    def get_tunnel_endpoints(self, agent_id: str, endpoint_type: str):
        return self.request(Route("GET", "/ch/v1/agent/{}/ngrok_tunnels/{}/", agent_id, endpoint_type))

    def get_tunnel(self, tunnel_id: str):
        return self.request(Route("GET", "/ch/v1/tunnels/{}/", tunnel_id))

    def get_tunnels(self, agent_id: str, show_choices: bool = False):
        return self.request(Route("GET", "/ch/v1/agent/{}/dd_tunnels/", agent_id, choices=show_choices))

    def create_tunnel(self, agent_id: str, **data):
        return self.request(Route("POST", "/ch/v1/agent/{}/dd_tunnels/", agent_id), json=data)

    def patch_tunnel(self, tunnel_id: str, **data):
        return self.request(Route("PATCH", "/ch/v1/tunnels/{}/", tunnel_id), json=data)

    def activate_tunnel(self, tunnel_id: str):
        return self.request(Route("POST", "/ch/v1/tunnels/{}/activate/", tunnel_id))

    def deactivate_tunnel(self, tunnel_id: str):
        return self.request(Route("POST", "/ch/v1/tunnels/{}/deactivate/", tunnel_id))

    def delete_tunnel(self, tunnel_id: str):
        return self.request(Route("DELETE", "/ch/v1/tunnels/{}/", tunnel_id))


    def fetch_token(self):
        """Fetch a temporary token from the cloud API.

        By default, this uses the username and password set in the client, and will fail if this is not set.

        You can override this to implement a custom token refresher, e.g. using the device agent websocket connection.

        This must return a tuple of (token, expires_at, agent_id).
        """
        if not (self.username or self.password):
            raise RuntimeError("Must have username and password set since access token has expired.")

        logging.info("Logging in...")
        session = requests.Session()

        login_url = f"{self.base_url}/accounts/login/"

        session.get(login_url)
        login_data = dict(
            login=self.username,
            password=self.password,
            csrfmiddlewaretoken=session.cookies.get('csrftoken'),
            next='/'
        )
        res = session.post(login_url, data=login_data, headers=dict(Referer=login_url))

        # bit of a hack... don't know a better way? Two-Factor is the title on the page...
        if "Two-Factor" in res.text:
            print("Your account has 2FA enabled. It is recommended to instead use `doover configure_token` "
                  "and use a long-lived token, otherwise you will have to 2FA authenticate every 20min.\n"
                  "Quit and run that command, or supply your 2FA code to authenticate now.\n")

            token = input("Please enter your 2FA token: ")
            twofa_data = dict(
                csrfmiddlewaretoken=session.cookies.get('csrftoken'),
                code=token,
            )
            twofa_url = f"{self.base_url}/accounts/2fa/authenticate/"
            res = session.post(twofa_url, data=twofa_data, headers=dict(Referer=twofa_url))
            if res.status_code != 200:
                raise RuntimeError("Failed to authenticate 2FA.")

        res = session.get(f"{self.base_url}/ch/v1/get_temp_token/")

        try:
            data = res.json()
        except requests.exceptions.JSONDecodeError:
            print(res.text)
            raise RuntimeError("Failed to get temporary token. Login failed.")

        # FIXME: can these expire in UTC?
        difference = timedelta(seconds=float(data["valid_until"]) - float(data["current_time"]))
        expires_at = datetime.utcnow() + difference
        return data["token"], expires_at, data["agent_id"]

    def login(self):
        token, expires_at, agent_id = self.fetch_token()
        self._set_login_data(token, expires_at, agent_id)

    def _set_login_data(self, token, expires_at, agent_id):
        self.access_token = AccessToken(token=token, expires_at=expires_at)
        self.agent_id = agent_id
        self.update_headers()

        logging.info(
            f"Successfully logged in and set token to expire "
            f"in {int((expires_at - datetime.utcnow()).total_seconds()/60)}min..."
        )
        try:
            self.login_callback()
        except Exception as e:
            print(f"failed to call callback: {e}")
            pass



import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional

from .element import Element
from .misc import Colour, Option
from ..utils import call_maybe_async

if TYPE_CHECKING:
    from .manager import UIManager

log = logging.getLogger(__name__)


class NotSet:
    pass


class Interaction(Element):
    type = "uiInteraction"

    def __init__(
        self, name: str, display_name: str = None, current_value: Any = NotSet,
        default: Any = None, callback=None, transform_check=None, show_activity: Optional[bool] = None, **kwargs
    ):
        super().__init__(name, display_name, **kwargs)
        self._current_value = current_value
        self._default_value = default or kwargs.pop("default_val", None)

        self._manager: Optional["UIManager"] = None

        if callback:
            self.callback = callback
        if transform_check:
            self.transform_check = transform_check

        ## Handle default value 'statelessly' now in current_value property instead of here
        # if self._current_value in (None, NotSet) and self._default_value is not None:
        #     print(f"Coercing {self.name} to default value {self._default_value}")
        #     self.coerce(self._default_value)

        self.show_activity = show_activity

    @property
    def current_value(self):
        if self._current_value is NotSet and self._default_value is not None:
            return self._default_value
        return self._current_value if self._current_value is not NotSet else None
    
    @current_value.setter
    def current_value(self, new_val):
        ## Store all datetime objects as epoch seconds internally
        if isinstance(new_val, datetime):
            new_val = int(new_val.timestamp())
        self._current_value = new_val

    def _json_safe_current_value(self):
        result = self.current_value
        if isinstance(result, datetime):
            return int(result.timestamp())
        return result

    def callback(self, new_value: Any):
        return

    def transform_check(self, new_value: Any) -> Any:
        """Transform and check a new value before setting it.

        This is called before any callback functions are invoked.

        By default, this will replace any None values with a set default value.
        """
        if new_value is None and self._default_value is not None:
            return self._default_value
        else:
            return new_value

    async def _handle_new_value(self, new_value: Any):
        try:
            new_value = self.transform_check(new_value)
        except Exception as e:
            log.error(f"Error transforming value for {self.name}: {e}")
            return

        self.current_value = new_value
        await call_maybe_async(self.callback, new_value)
        # try:
        #     self.callback(new_value)
        # except Exception as e:
        #     log.error(f"Error in callback for {self.name}: {e}")

    def coerce(self, value: Any, critical: bool = False):
        if critical and self._manager and value != self.current_value:
            self._manager._has_critical_interaction_pending = True

        self.current_value = value

    def to_dict(self):
        res = super().to_dict()
        if self._current_value is not NotSet:
            res['currentValue'] = self._json_safe_current_value()
        if self.show_activity is not None:
            res['showActivity'] = self.show_activity
        return res


class Action(Interaction):
    type = "uiAction"

    def __init__(
        self, name: str, display_name: str, colour: Colour = Colour.blue, requires_confirm: bool = True, **kwargs
    ):
        super().__init__(name, display_name, **kwargs)
        self.colour = colour
        self.requires_confirm = requires_confirm

    def to_dict(self):
        result = super().to_dict()
        result['colour'] = str(self.colour)
        result['requiresConfirm'] = self.requires_confirm
        return result


class WarningIndicator(Interaction):
    type = "uiWarningIndicator"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.can_cancel = kwargs.pop("can_cancel", True)

    def to_dict(self):
        result = super().to_dict()
        result["can_cancel"] = self.can_cancel
        return result


class HiddenValue(Interaction):
    type = "uiHiddenValue"

    def __init__(self, name, **kwargs):
        super().__init__(name, display_name=None, **kwargs)

    def to_dict(self):
        return {
            "name": self.name,
            "type": self.type,
        }


class SlimCommand(Interaction):
    type = "uiStateCommand"


class StateCommand(SlimCommand):
    type = "uiStateCommand"

    def __init__(self, name: str, display_name: str = None, user_options: list[Option] = None, **kwargs):
        super().__init__(name, display_name, **kwargs)

        # A list of doover_ui_element's
        self.user_options = []
        self.add_user_options(*user_options)

    def to_dict(self):
        result = super().to_dict()
        result["userOptions"] = {o.name: o.to_dict() for o in self.user_options}
        return result

    def add_user_options(self, *option: Option):
        for o in option:
            # still support legacy dict passing of option values.
            if isinstance(o, Option):
                self.user_options.append(o)
            elif isinstance(o, dict):
                self.user_options.append(Option.from_dict(o))

    add_user_option = add_user_options


class Slider(Interaction):
    type = "uiSlider"

    def __init__(self, name: str, display_name: str = None, min_val: int = 0, max_val: int = 100, step_size: float = 0.1,
                 dual_slider: bool = True, inverted: bool = True, icon: Optional[str] = None, colours: Optional[str] = None, **kwargs
        ):
        super().__init__(name, display_name, **kwargs)
        self.min_val = min_val
        self.max_val = max_val
        self.step_size = step_size
        self.dual_slider = dual_slider
        self.inverted = inverted
        self.icon = icon
        self.colours = colours
        
    def to_dict(self):
        result = super().to_dict()
        result["min"] = self.min_val
        result["max"] = self.max_val
        result["stepSize"] = self.step_size
        result["dualSlider"] = self.dual_slider
        result["isInverted"] = self.inverted
        result["icon"] = self.icon
        result["colours"] = self.colours
        return result


doover_ui_interaction = Interaction
doover_ui_action = Action
doover_ui_state_command = StateCommand
doover_ui_warning_indicator = WarningIndicator
doover_ui_hidden_value = HiddenValue
doover_ui_slider = Slider


def action(name: str, display_name: str = None, colour: Colour = Colour.blue, requires_confirm: bool = True, **kwargs):
    def decorator(func):
        func._ui_type = Action
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "colour": colour,
            "requires_confirm": requires_confirm,
            **kwargs
        }
        return func
    return decorator


def warning_indicator(name: str, display_name: str = None, can_cancel: bool = True, **kwargs):
    def decorator(func):
        func._ui_type = WarningIndicator
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "can_cancel": can_cancel,
            **kwargs,
        }
        return func
    return decorator


def state_command(name: str, display_name: str = None, user_options: list[Option] = None, **kwargs):
    def decorator(func):
        func._ui_type = StateCommand
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "user_options": user_options,
            **kwargs,
        }
        return func
    return decorator


def hidden_value(name: str, **kwargs):
    def decorator(func):
        func._ui_type = HiddenValue
        func._ui_kwargs = {
            "name": name,
            **kwargs,
        }
        return func
    return decorator


def slider(name: str, display_name: str = None, min_val: int = 0, max_val: int = 100, step_size: float = 0.1,
        dual_slider: bool = True, inverted: bool = True, icon: Optional[str] = None, **kwargs):
    def decorator(func):
        func._ui_type = Slider
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "min_val": min_val,
            "max_val": max_val,
            "step_size": step_size,
            "dual_slider": dual_slider,
            "inverted": inverted,
            "icon": icon,
            **kwargs,
        }
        return func
    return decorator



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ResponseHeader(_message.Message):
    __slots__ = ("success", "response_code", "message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_CODE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    response_code: int
    message: str
    def __init__(self, success: bool = ..., response_code: _Optional[int] = ..., message: _Optional[str] = ...) -> None: ...

class TestCommsRequest(_message.Message):
    __slots__ = ("message",)
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class TestCommsResponse(_message.Message):
    __slots__ = ("response_header", "response")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    response: str
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., response: _Optional[str] = ...) -> None: ...

class TunnelRequest(_message.Message):
    __slots__ = ("address", "protocol", "timeout", "username", "password", "domain", "remote_addr", "allow_cidrs")
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    PROTOCOL_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    PASSWORD_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    REMOTE_ADDR_FIELD_NUMBER: _ClassVar[int]
    ALLOW_CIDRS_FIELD_NUMBER: _ClassVar[int]
    address: str
    protocol: str
    timeout: int
    username: str
    password: str
    domain: str
    remote_addr: str
    allow_cidrs: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, address: _Optional[str] = ..., protocol: _Optional[str] = ..., timeout: _Optional[int] = ..., username: _Optional[str] = ..., password: _Optional[str] = ..., domain: _Optional[str] = ..., remote_addr: _Optional[str] = ..., allow_cidrs: _Optional[_Iterable[str]] = ...) -> None: ...

class TunnelResponse(_message.Message):
    __slots__ = ("address", "url")
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    address: str
    url: str
    def __init__(self, address: _Optional[str] = ..., url: _Optional[str] = ...) -> None: ...

class OpenTunnelRequest(_message.Message):
    __slots__ = ("tunnel",)
    TUNNEL_FIELD_NUMBER: _ClassVar[int]
    tunnel: TunnelRequest
    def __init__(self, tunnel: _Optional[_Union[TunnelRequest, _Mapping]] = ...) -> None: ...

class CloseTunnelRequest(_message.Message):
    __slots__ = ("address", "url")
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    address: str
    url: str
    def __init__(self, address: _Optional[str] = ..., url: _Optional[str] = ...) -> None: ...

class CloseAllTunnelsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetTunnelRequest(_message.Message):
    __slots__ = ("address", "url")
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    address: str
    url: str
    def __init__(self, address: _Optional[str] = ..., url: _Optional[str] = ...) -> None: ...

class ListTunnelsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class OpenTunnelResponse(_message.Message):
    __slots__ = ("response_header", "success", "url")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    success: bool
    url: str
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., success: bool = ..., url: _Optional[str] = ...) -> None: ...

class CloseTunnelResponse(_message.Message):
    __slots__ = ("response_header", "success")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    success: bool
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., success: bool = ...) -> None: ...

class CloseAllTunnelsResponse(_message.Message):
    __slots__ = ("response_header", "success")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    success: bool
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., success: bool = ...) -> None: ...

class GetTunnelResponse(_message.Message):
    __slots__ = ("response_header", "tunnel")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    TUNNEL_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    tunnel: TunnelResponse
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., tunnel: _Optional[_Union[TunnelResponse, _Mapping]] = ...) -> None: ...

class ListTunnelsResponse(_message.Message):
    __slots__ = ("response_header", "tunnels")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    TUNNELS_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    tunnels: _containers.RepeatedCompositeFieldContainer[TunnelResponse]
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., tunnels: _Optional[_Iterable[_Union[TunnelResponse, _Mapping]]] = ...) -> None: ...

class SyncTunnelsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SyncTunnelsResponse(_message.Message):
    __slots__ = ("response_header", "success")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    success: bool
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., success: bool = ...) -> None: ...



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class StreamSettings(_message.Message):
    __slots__ = ("fps", "scale")
    FPS_FIELD_NUMBER: _ClassVar[int]
    SCALE_FIELD_NUMBER: _ClassVar[int]
    fps: int
    scale: str
    def __init__(self, fps: _Optional[int] = ..., scale: _Optional[str] = ...) -> None: ...

class ResponseHeader(_message.Message):
    __slots__ = ("success", "response_code", "message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_CODE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    response_code: int
    message: str
    def __init__(self, success: bool = ..., response_code: _Optional[int] = ..., message: _Optional[str] = ...) -> None: ...

class TestCommsRequest(_message.Message):
    __slots__ = ("message",)
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class TestCommsResponse(_message.Message):
    __slots__ = ("response_header", "response")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    response: str
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., response: _Optional[str] = ...) -> None: ...

class SnapshotRequest(_message.Message):
    __slots__ = ("rtsp_uri", "snapshot_type", "snapshot_length", "output_channel", "camera_name", "stream_settings")
    RTSP_URI_FIELD_NUMBER: _ClassVar[int]
    SNAPSHOT_TYPE_FIELD_NUMBER: _ClassVar[int]
    SNAPSHOT_LENGTH_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_CHANNEL_FIELD_NUMBER: _ClassVar[int]
    CAMERA_NAME_FIELD_NUMBER: _ClassVar[int]
    STREAM_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    rtsp_uri: str
    snapshot_type: str
    snapshot_length: int
    output_channel: str
    camera_name: str
    stream_settings: StreamSettings
    def __init__(self, rtsp_uri: _Optional[str] = ..., snapshot_type: _Optional[str] = ..., snapshot_length: _Optional[int] = ..., output_channel: _Optional[str] = ..., camera_name: _Optional[str] = ..., stream_settings: _Optional[_Union[StreamSettings, _Mapping]] = ...) -> None: ...

class SnapshotResponse(_message.Message):
    __slots__ = ("response_header", "response")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    response: str
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., response: _Optional[str] = ...) -> None: ...

class IsSnapshotRunningRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class IsSnapshotRunningResponse(_message.Message):
    __slots__ = ("response_header", "is_running")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    IS_RUNNING_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    is_running: bool
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., is_running: bool = ...) -> None: ...



"""
This is the base definition for a tiny lambda function
Which is run in response to messages processed in Doover's 'Channels' system.
This base class is designed to be overridden by a user defined class

In the doover_config.json file we have defined some of these subscriptions
These are under 'processor_deployments' > 'tasks'
"""

import logging
import sys
import time

from typing import Any

from ...cloud.api import Client, Message

from ...ui import UIManager

# use the root logger since we want to pipe these logs to a channel.
log = logging.getLogger()


class LogHandler(logging.NullHandler):
    def __init__(self, *args, **kwargs):
        self.logs = []
        super().__init__(*args, **kwargs)

    def handle(self, record):
        if record.levelno < self.level:
            return

        fmt = self.format(record)
        self.logs.append(fmt)

    def emit(self, record):
        self.handle(record)

    def get_logs(self):
        return "\n".join(self.logs)


class ProcessorBase:
    def __init__(self, **kwargs):

        self.agent_id: str = kwargs["agent_id"]
        self.access_token: str = kwargs["access_token"]
        self.log_channel_id: str = kwargs["log_channel"]
        self.task_id: str = kwargs["task_id"]

        self.api: Client = Client(token=self.access_token, base_url=kwargs["api_endpoint"])
        self.ui_manager: UIManager = UIManager(self.agent_id, self.api)
        
        self._log_handler = LogHandler()
        log.addHandler(self._log_handler)
        log.setLevel(level=logging.INFO)

        self.agent_id: str = kwargs["agent_id"]
        self.log_channel_id: str = kwargs["log_channel"]
        self.task_id: str = kwargs["task_id"]

        self.deployment_config: dict[str, Any] = kwargs["agent_settings"].get("deployment_config", {})
        self.package_config: dict[str, Any] = kwargs.get("package_config", {})

        try:
            if kwargs["msg_obj"] is None:
                raise KeyError
            self.message = Message(client=self.api, data=kwargs["msg_obj"], channel_id=None)
        except KeyError:
            self.message = None

        ### kwarg
        #     'agent_id' : The Doover agent id invoking the task e.g. '9843b273-6580-4520-bdb0-0afb7bfec049'
        #     'access_token' : A temporary token that can be used to interact with the Doover API .e.g 'ABCDEFGHJKLMNOPQRSTUVWXYZ123456890',
        #     'api_endpoint' : The API endpoint to interact with e.g. "https://my.doover.com",
        #     'package_config' : A dictionary object with configuration for the task - as stored in the task channel in Doover,
        #     'msg_obj' : A dictionary object of the msg that has invoked this task,
        #     'task_id' : The identifier string of the task channel used to run this processor,
        #     'log_channel' : The identifier string of the channel to publish any logs to
        #     'agent_settings' : {
        #       'deployment_config' : {} # a dictionary of the deployment config for this agent
        #     }

    def setup(self):
        return NotImplemented

    def close(self):
        return NotImplemented

    def execute(self):
        """This function is invoked after the singleton instance is created."""
        start_time = time.time()
        log.info(f"Initialising processor task for task channel {self.task_id}")
        log.info(f"Started at {start_time}.")

        try:
            self.import_modules()
            self.setup()

            try:
                self.process()
            except Exception as e:
                log.error(f"ERROR attempting to process message: {e} ", exc_info=e)

        except Exception as e:
            log.error(f"ERROR attempting to initialise process: {e}", exc_info=e)

        try:
            self.close()
        except Exception as e:
            log.error(f"ERROR attempting to close process: {e} ", exc_info=e)

        end_time = time.time()
        log.info(f"Finished at {end_time}. Process took {end_time - start_time} seconds.")

        if self._log_handler.get_logs() and self.log_channel_id is not None:
            self.api.publish_to_channel(self.log_channel_id, self._log_handler.get_logs())

    def process(self):
        return NotImplemented
    
        ## Do some logic
    
        ## Optionally update the UI at the end
        # self.ui_manager.push()

    @staticmethod
    def import_modules():
        """Attempt to delete any loaded pydoover modules that persist across lambdas."""
        if 'pydoover' in sys.modules:
            del sys.modules['pydoover']
        try:
            del pydoover
        except:
            pass
        try:
            del pd
        except:
            pass

    def get_agent_config(self, filter_key: str = None):
        if not self.deployment_config:
            return None
        if filter_key:
            return self.deployment_config.get(filter_key)
        return self.deployment_config

    def fetch_channel(self, channel_id: str):
        return self.api.get_channel(channel_id)

    def fetch_channel_named(self, channel_name: str):
        return self.api.get_channel_named(channel_name, self.agent_id)



import json, time, csv
from datetime import datetime
from typing import Any


class Message:

    def __init__(self, client, data, channel_id=None, agent_id=None, channel_name=None):
        
        self.id = None
        self.timestamp = None

        self.client = client
        self.channel_id = channel_id
        self.agent_id = agent_id
        self.channel_name = channel_name
        self._payload = None

        if data is not None:
            self._from_data(data)

    def __repr__(self):
        if self._payload is not None:
            return f"<Message message_id={self.id}, payload={self._payload}>"
        return f"<Message message_id={self.id}>"

    def _from_data(self, data: dict[str, Any]):
        # {'agent': '9fb5d629-ce7f-4b08-b17a-c267cbcd0427', 'message': 'a7b493dd-4577-4f81-ac3d-f1b3be680b12', 'type': 'base', 'timestamp': 1715646840.250541}
        self.id = data.get("message", None)
        self.agent_id = data.get("agent", None)
        self.channel_name = data.get("channel_name", None)
        self.timestamp = data.get("timestamp", None)

        if not self.channel_id:
            self.channel_id = data.get("channel")

        self._payload = data.get("payload")

    def to_dict(self):
        return {
            "message": self.id,
            "agent": self.agent_id,
            "timestamp": self.timestamp,
            "channel": self.channel_id,
            "channel_name": self.channel_name,
            "payload": self._payload
        }

    def update(self):
        data = self.client._get_message_raw(self.channel_id, self.id)
        self._from_data(data)

    def delete(self):
        self.client._delete_message_raw(self.channel_id, self.id)

    def fetch_payload(self):
        if self._payload is not None:
            return self._payload

        data = self.client._get_message_raw(self.channel_id, self.id)
        self._payload = json.loads(data["payload"])
        return self._payload

    def get_age(self):
        return time.time() - self.timestamp

    def get_timestamp(self):
        return datetime.fromtimestamp(self.timestamp)

    @staticmethod
    def from_csv_export(client, csv_file_path):
        
        messages = []

        # Open and read the CSV file using the csv module
        with open(csv_file_path, 'r', newline='') as file:
            reader = csv.DictReader(file)  # Use DictReader to handle headers

            for row in reader:
                # Extract data from the row
                key = row['Key']
                timestamp = row['Timestamp (UTC)']
                channel_name = row['Channel']
                channel_id = row['Channel ID']
                agent_name = row['Agent']
                agent_id = row['Agent ID']
                payload = row['Payload']

                # Convert timestamp to UTC epoch timestamp
                timestamp = datetime.fromisoformat(timestamp).timestamp()

                # Create a Message instance
                message = Message(
                    client,
                    data=None,
                    channel_id=channel_id,
                    agent_id=agent_id,
                    channel_name=channel_name
                )

                message.id = key
                message.timestamp = timestamp
                message._payload = json.loads(payload)

                messages.append(message)

        # Sort the messages by timestamp
        messages.sort(key=lambda x: x.timestamp)

        return messages



#!/usr/bin/env python3

import asyncio
import logging
import time
from typing import Dict

from pydoover.utils import apply_async_kalman_filter

class PowerManager:
    def __init__(self, app):
        self.app = app
        self.start_time = None
        self.config = {}
        self.scheduled_sleep_time = None ## Secs that the system is scheduled to sleep for
        self.scheduled_goto_sleep_time = None ## Time that the system is scheduled to sleep

        self.last_voltage = None
        self.last_voltage_time = None
        self.voltage_update_interval = 5

        self.request_shutdown_hooks = [] # Request shutdown hooks will be run and if all return True, the system will shutdown

        self.pre_shutdown_hooks = [] # Pre shutdown hooks will be run when the system schedules a shutdown
        self.shutdown_hooks = [] # Shutdown hooks will be run when the system is about to shutdown

    def register_request_shutdown_hook(self, hook):
        self.request_shutdown_hooks.append(hook)

    def register_pre_shutdown_hook(self, hook):
        self.pre_shutdown_hooks.append(hook)

    def register_shutdown_hook(self, hook):
        self.shutdown_hooks.append(hook)

    async def fetch_config(self):
        self.config = await self.app.get_config_async(key_filter="POWER_MANAGEMENT", wait=True)
        if not self.config or not self.config.get("VOLTAGE_SLEEP_MINUTES"):
            logging.info("No configuration found for PowerManager. Disabling Power Management.")

    def is_active(self):
        return bool(self.config)

    async def update_voltage(self):
        # Only update the voltage every voltage_update_interval seconds
        if self.last_voltage_time is None or asyncio.get_event_loop().time() - self.last_voltage_time > self.voltage_update_interval:
            # self.last_voltage = None
            try:
                self.last_voltage = await self.get_system_voltage()
            except Exception as e:
                logging.error(f"Error fetching system voltage: {e}")

            if self.last_voltage is not None:
                self.last_voltage = round(self.last_voltage, 2)
                self.last_voltage_time = asyncio.get_event_loop().time()
            logging.info(f"Filtered system voltage: {self.last_voltage}")

    @apply_async_kalman_filter(
        process_variance=0.05,
        outlier_threshold=0.5,
    )
    async def get_system_voltage(self) -> float:
        # Get the current system voltage
        voltage = await self.app.platform_iface.get_system_voltage_async()
        return voltage

    def get_sleep_time(self) -> int:
        sorted_config = sorted(self.config.get("VOLTAGE_SLEEP_MINUTES", {}).items())
        voltage = self.last_voltage
        if voltage is None:
            return None
        for threshold, sleep_time in sorted_config:
            if isinstance(threshold, str):
                threshold = float(threshold)
            if voltage <= threshold:
                return sleep_time * 60 # Convert minutes to seconds
        return None
    
    def get_min_awake_time(self) -> int:
        abs_min_awake_time = 30 ## The floor value for the minimum awake time
        min_awake_time = self.config.get("MIN_AWAKE_SECONDS", 60)
        if isinstance(min_awake_time, int):
            return min_awake_time
        
        ## Assume min_awake_time is a dict
        # If the min_awake_time is a dict, get the value for the current voltage
        try:
            voltage = self.last_voltage
            if voltage is None:
                return abs_min_awake_time
            for threshold, time in sorted(min_awake_time.items()):
                if isinstance(threshold, str):
                    threshold = float(threshold)
                if voltage <= threshold:
                    return time
        except Exception as e:
            logging.warning(f"Error getting min_awake_time: {e}")
        return abs_min_awake_time        

    def get_awake_time(self) -> int:
        if self.start_time is None:
            self.start_time = asyncio.get_event_loop().time()
        return int(asyncio.get_event_loop().time() - self.start_time)

    async def maybe_schedule_sleep(self, sleep_time: int, time_till_sleep: int = 20):
        if self.scheduled_goto_sleep_time is not None:
            if self.get_time_till_sleep() is not None:
                logging.warning(f"Time till sleep: {self.get_time_till_sleep()}")
            return

        if self.get_awake_time() < (self.get_min_awake_time() - time_till_sleep):
            time_till_sleep = self.get_min_awake_time() - self.get_awake_time()
            logging.info("Minimum awake time not met: {} seconds to go".format(time_till_sleep))
            return

        if not self.shutdown_permitted():
            logging.info("Scheduling of shutdown not yet permitted by application. Waiting...")
            return
        
        immunity_time = await self.get_immunity_time()
        if immunity_time is not None:
            logging.info(f"Device immune to shutdown for {immunity_time} seconds.")
            return

        logging.info(f"Scheduling sleep of {sleep_time} secs in {time_till_sleep} secs.")
        self.scheduled_goto_sleep_time = asyncio.get_event_loop().time() + time_till_sleep
        self.scheduled_sleep_time = sleep_time
        await self.run_pre_shutdown_hooks(time_till_sleep, sleep_time)

    def get_time_till_sleep(self):
        if self.scheduled_goto_sleep_time is None:
            return None
        return int(self.scheduled_goto_sleep_time - asyncio.get_event_loop().time())

    def is_ready_to_sleep(self):
        if self.scheduled_goto_sleep_time is None:
            return False
        return asyncio.get_event_loop().time() >= self.scheduled_goto_sleep_time

    def shutdown_permitted(self):
        for hook in self.request_shutdown_hooks:
            if not hook():
                return False
        return True
    
    async def get_immunity_time(self):
        immunity_secs = await self.app.platform_iface.get_immunity_seconds_async()
        if immunity_secs is not None and immunity_secs <= 1:
            immunity_secs = None
        return immunity_secs

    async def schedule_next_startup(self):
        if self.scheduled_sleep_time is None:
            self.scheduled_sleep_time = self.get_sleep_time()
        await self.app.platform_iface.schedule_startup_async(self.scheduled_sleep_time)

    async def assess_power(self):
        """
        Monitor system voltage, determine sleep time, and handle shutdown.
        """
        if not self.is_active():
            return
        
        if self.start_time is None:
            self.start_time = asyncio.get_event_loop().time()

        # Update the system voltage
        await self.update_voltage()

        # If the system is already scheduled to sleep, check if it's time to sleep
        if self.is_ready_to_sleep():
            await self.go_to_sleep()
            return                

        # Determine the sleep time from the config & current voltage
        sleep_time = self.get_sleep_time()
        if sleep_time is None:
            return

        # Attempt to schedule the sleep
        await self.maybe_schedule_sleep(sleep_time)

    async def go_to_sleep(self):
        """
        Put the system to sleep.
        """
        logging.warning("Putting system to sleep...")
        
        ## Run shutdown hooks
        for hook in self.shutdown_hooks:
            await self.run_hook(hook)

        ## schedule the next startup
        await self.schedule_next_startup()

        ## Put the system to sleep
        await self.app.platform_iface.shutdown_async()

        ## Cleanly disconnect the device comms and then wait for sleep
        try:
            self.app.close_app(with_delay=120)
        except Exception as e:
            logging.error(f"Error closing device application for shutdown: {e}")

        ## Wait for the system to shutdown
        time.sleep(120)

    async def run_pre_shutdown_hooks(self, time_till_sleep, sleep_time):
        for hook in self.pre_shutdown_hooks:
            await self.run_hook(
                hook,
                kwargs={
                    "time_till_sleep": time_till_sleep,
                    "sleep_time": sleep_time
                },
            )

    async def run_hook(self, hook, kwargs={}):
        try:
            ## If the hook is a coroutine, await it
            if asyncio.iscoroutinefunction(hook):
                await hook(**kwargs)
            else:
                hook(**kwargs)
        except Exception as e:
            logging.error(f"Error running hook {hook}: {e}")

    async def setup(self):
        logging.info("Setting up PowerManager...")
        await self.fetch_config()
        if not self.is_active():
            return
        
        ## Attempt 3 times to get a non-None voltage
        for i in range(3):
            if self.last_voltage is None:
                await self.update_voltage()
                await asyncio.sleep(0.1)
            else:
                break

    async def main_loop(self):
        await self.assess_power()



# Test Function
if __name__ == "__main__":

    # Simulate the app's interface with required stubs
    class MockPlatformInterface:
        async def get_system_voltage_async(self):
            if not hasattr(self, "start_time"):
                self.start_time = asyncio.get_event_loop().time()
            run_time = asyncio.get_event_loop().time() - self.start_time

            if not hasattr(self, "run_count"):
                self.run_count = 0
            self.run_count += 1

            ## Test Voltages - A steadily decreasing voltage from 13.8 to 11.8 V
            voltage = 13.8 - run_time / 50
            if voltage < 11.8:
                voltage = 11.8

            ## Add a random noise to the voltage (up to 3v)
            import random
            voltage += random.random() * 3
            
            ## Every 5th iteration, simulate a voltage drop
            if self.run_count % 6 < 2:
                voltage -= 5

            ## Every 6th iteration return None
            if self.run_count % 7 < 2:
                voltage = None
            else:
                voltage = round(voltage, 2)

            logging.debug(f"Simulated voltage: {voltage}")
            return voltage

        async def schedule_startup_async(self, sleep_time):
            logging.error(f"Startup scheduled after {sleep_time} seconds.")

        async def shutdown_async(self):
            logging.error("System shutting down...")

    class MockApp:
        def __init__(self):
            self.platform_iface = MockPlatformInterface()

        async def get_config_async(self, key_filter=None, wait=True):
            # Simulated configuration for testing
            if key_filter == "POWER_MANAGEMENT":
                return {
                    "VOLTAGE_SLEEP_MINUTES": {
                        12.7: 15,  # 15 mins of sleep if voltage <= 12.7
                        12.2: 60,  # 1 hour of sleep if voltage <= 12.2
                        11.8: 240,   # 4 hours sleep if voltage <= 11.8
                    },
                    # "MIN_AWAKE_SECONDS": 45,  # Minimum awake time in seconds
                    "MIN_AWAKE_SECONDS" : {
                        12.7: 120,
                        12.2: 60,
                        11.8: 40,
                    }
                }
            return {}

    # Define pre-shutdown hooks
    def sync_pre_shutdown_hook(**kwargs):
        logging.info("Running synchronous pre-shutdown hook. {}".format(kwargs))

    async def async_pre_shutdown_hook(**kwargs):
        logging.info("Running asynchronous pre-shutdown hook. {}".format(kwargs))
        await asyncio.sleep(1)

    def sync_shutdown_hook(**kwargs):
        logging.info("Running synchronous shutdown hook. {}".format(kwargs))

    async def async_shutdown_hook(**kwargs):
        logging.info("Running asynchronous shutdown hook. {}".format(kwargs))
        await asyncio.sleep(1)

    # Main testing function
    async def main():
        logging.basicConfig(level=logging.DEBUG)

        # Instantiate the mock app and PowerManager
        app = MockApp()
        power_manager = PowerManager(app)

        # Register pre-shutdown hooks
        power_manager.register_pre_shutdown_hook(sync_pre_shutdown_hook)
        power_manager.register_pre_shutdown_hook(async_pre_shutdown_hook)

        # Register shutdown hooks
        power_manager.register_shutdown_hook(sync_shutdown_hook)
        power_manager.register_shutdown_hook(async_shutdown_hook)

        # Setup the PowerManager
        await power_manager.setup()

        # Simulate a short monitoring loop for testing
        while True:
            await power_manager.main_loop()
            await asyncio.sleep(2)


    asyncio.run(main())


from .doover_docker import app_base, app_manager, run_app, deployment_config_manager
from .camera import CameraInterface, Camera, CameraManager, camera_iface
from .device_agent import DeviceAgentInterface, device_agent_iface
from .platform import PlatformInterface, PulseCounter, platform_iface, pulse_counter
from .tunnel import TunnelInterface
from .power_manager import PowerManager
from .location_manager import LocationManager


import asyncio
import json
import logging
import time
import json

from collections.abc import Iterable
from typing import Optional, Callable, Any

import grpc

from .grpc_stubs import platform_iface_pb2, platform_iface_pb2_grpc
from ..grpc_interface import GRPCInterface
from ...utils import maybe_async, call_maybe_async, deprecated
from ...cli.decorators import command as cli_command, annotate_arg

class PulseCounter:
    def __init__(
        self,
        plt_iface: "PlatformInterface",
        pin: int,
        edge: str = "rising",
        callback: Callable[[int, bool, int, int, str], Any] = None,
        rate_window_secs: int = 60,
        auto_start: bool = True,
    ):
        self.platform_iface = plt_iface
        self.pin = pin
        self.edge = edge
        self.callback = callback
        self.rate_window_secs = rate_window_secs
        self.count = 0

        self.start_time = time.time()
        self.pulse_grace_period = 0.2  # Need to ignore pulses for a short period after starting
        self.pulse_timestamps = []
        
        self.recieving_pulses = False
        self.recieving_events = False

        if auto_start:
            self.start_listener_pulses()

    @deprecated("Replaced with PulseCounter.start_listener_pulses")
    def start_listener(self):
        self.start_listener_pulses()
        
    def start_listener_pulses(self):
        
        # Check each pulse counter is only used for a single type of pulse
        if self.recieving_events:
            logging.error("Using a pulse counter for both pulses and offline events")
            return
        self.recieving_pulses = True
        
        self.start_time = time.time()
        self.platform_iface.start_di_pulse_listener(self.pin, self.receive_pulse, edge=self.edge, start_count=self.count)
        
    def update_events(self):
        
        # Check each pulse counter is only used for a single type of pulse
        if self.recieving_pulses:
            logging.error("Using a pulse counter for both pulses and offline events")
            return
        self.recieving_events = True
        
        self.recieve_events(self.platform_iface.get_di_events(self.pin, self.edge, include_system_events=True))
        
    def add_existing_events(self, time_stamps):
        
        # Check each pulse counter is only used for a single type of pulse
        if self.recieving_pulses:
            logging.error("Using a pulse counter for both pulses and offline events")
            return
        self.recieving_events = True
        
        self.pulse_timestamps += time_stamps
        self.count = len(self.pulse_timestamps)

    async def receive_pulse(self, di, di_value, dt_secs, counter, edge):
        
        # Check each pulse counter is only used for a single type of pulse
        if self.recieving_events:
            logging.error("Using a pulse counter for both pulses and offline events")
            return
        self.recieving_pulses = True
        
        if time.time() - self.start_time < self.pulse_grace_period:
            logging.info("Ignoring pulse on di=" + str(di) + " with dt=" + str(dt_secs) + "s")
            return
        logging.info("Received pulse on di=" + str(di) + " with dt=" + str(dt_secs) + "s")
        self.count += 1
        self.pulse_timestamps += [time.time()]
        if self.callback is not None:
            await call_maybe_async(self.callback, self.pin, di_value, dt_secs, self.count, edge)
            
    def recieve_events(self, events):
        
        # Check each pulse counter is only used for a single type of pulse
        if self.recieving_pulses:
            logging.error("Using a pulse counter for both pulses and offline events")
            return
        self.recieving_events = True
                
        for event in events:
            edge = ""
            di_value = 0
            if event.event == "DI_R":
                di_value = 1
                edge = "rising"
            elif event.event == "DI_F":
                edge = "falling"
            else: # Could be a system event
                self.handle_system_event(event)
                continue
            
            timestamp = event.time / 1000 or time.time()
            dt_secs = 0
            if len(self.pulse_timestamps) > 0:
                if timestamp <= self.pulse_timestamps[-1] + 0.01:
                    logging.warning(f"Ignoring old event on di={event.pin} t={timestamp} latest event: {self.pulse_timestamps[-1]}")
                    continue
                dt_secs = timestamp - self.pulse_timestamps[-1]
            logging.info(f"Received event on di={event.pin} with t=" + str(dt_secs) + "s")
            self.count += 1
            self.pulse_timestamps += [timestamp]
            if self.callback is not None:
                self.callback(self.pin, di_value, dt_secs, timestamp, self.count, edge)
    
    def handle_system_event(self, event):
        pass

    @deprecated("Use get_pulses_in_window to not damage record of pulses/events")
    def clean_pulse_timestamps(self):
        if len(self.pulse_timestamps) == 0:
            return

        ## Remove timestamps older than the rate window
        while len(self.pulse_timestamps) > 0 and self.pulse_timestamps[0] < time.time() - self.rate_window_secs:
            self.pulse_timestamps.pop(0)
            
    def get_pulses_in_window(self):
        if len(self.pulse_timestamps) == 0:
            return []

        pulses = []
        ## Remove timestamps older than the rate window
        for timestamp in self.pulse_timestamps:
            if timestamp > self.pulse_timestamps[-1] - self.rate_window_secs:
                pulses.append(timestamp)
        return pulses

    def set_rate_window(self, rate_window_secs):
        self.rate_window_secs = rate_window_secs

    def get_rate_window(self):
        return self.rate_window_secs

    def get_pulses_per_minute(self):
        pulses = self.get_pulses_in_window()
        return len(pulses) * 60 / self.rate_window_secs

    def set_counter(self, counter):
        self.count = counter

    def get_counter(self):
        return self.count
    


class PlatformInterface(GRPCInterface):
    stub = platform_iface_pb2_grpc.platformIfaceStub

    @annotate_arg("plt_uri", "the address that platform interface is running")
    def __init__(self, plt_uri: str = "localhost:50053", is_async: bool = False):
        super().__init__(plt_uri, is_async)
        self.pulse_counter_listeners = []

    def close(self):
        logging.info("Closing platform interface...")
        for listener in self.pulse_counter_listeners:
            listener.cancel()

    def process_response(self, stub_call: str, response, **kwargs):
        response = super().process_response(stub_call, response, **kwargs)

        try:
            response_field = kwargs.pop("response_field")
        except KeyError:
            return response

        res = getattr(response, response_field, None)
        if isinstance(res, Iterable) and not isinstance(res, str):  # don't iterate over strings
            res = list(res)

        if isinstance(res, list) and len(res) == 1:
            return res[0]

        return res

    
    def get_new_pulse_counter(
        self,
        di: int,
        edge: str = "rising",
        callback: Callable[[int, bool, int, int, str], Any] = None,
        rate_window_secs: int = 20,
        auto_start: bool = True
    ) -> PulseCounter:
        return PulseCounter(self, di, edge=edge, callback=callback, rate_window_secs=rate_window_secs, auto_start=auto_start)
    
    def get_new_event_counter(
        self,
        di: int,
        edge: str = "rising",
        callback: Callable[[int, bool, int, int, str], Any] = None,
        rate_window_secs: int = 20,
        auto_collect: bool = True
    ) -> PulseCounter:
        """Create a new Pulse Counter for counting events.
            counter = pltiface.get_new_event_counter(0, "rising")
            print(counter.get_counter())
            print(counter.get_pulses_per_minute())

        Parameters
        ----------
        di : int
            Pin number to check events for.
        edge : "rising" or "falling" or "both"
            The edge to listen to evenets on.
        callback : (pin, di_value, dt_secs, timestamp, count, edge) -> any
            Callback called when an event is processed.
        rate_window_secs : int = 20
            The size of window for which the rate of events is calculated.
        auto_collect : bool = True
            Whether to automatically collect the events from the platform interface.

        Returns
        -------
        PulseCounter object
        """
        counter = PulseCounter(self, di, edge=edge, callback=callback, rate_window_secs=rate_window_secs, auto_start=False)
        if auto_collect:
            counter.update_events()
        return counter

    def start_di_pulse_listener(self, di: int, callback, edge: str = "rising", start_count: int = 0):
        ## Callback should be a function that takes the following arguments:
        ## di, di_value, dt_secs, counter, edge

        listener = asyncio.create_task(self.recv_di_pulses(di, callback, edge=edge, start_count=start_count))
        self.pulse_counter_listeners.append(listener)
        listener.add_done_callback(self.pulse_counter_listeners.remove)
    
    async def recv_di_pulses(self, di: int, callback, edge: str = "rising", start_count: int = 0):
        counter = start_count
        active_callbacks = set()

        while True:
            try:
                # Setup the connection to the platform interface
                async with grpc.aio.insecure_channel(self.uri) as channel:
                    channel_stream = platform_iface_pb2_grpc.platformIfaceStub(channel).startPulseCounter(platform_iface_pb2.pulseCounterRequest(di=di))

                    while True:

                        response = await channel_stream.read()
                        if response is None or response == grpc.aio.EOF:
                            logging.info("pulseCounter for di=" + str(di) + " ended.")
                            break
                        
                        logging.debug("Received response from pulseCounter for di=" + str(di))

                        if hasattr(response, "dt_secs") and response.dt_secs is not None and response.dt_secs > 0:
                            ## Increment the counter
                            counter += 1
                            ## Call the callback function with the response
                            task = await call_maybe_async(callback, di, response.value, response.dt_secs, counter, edge, as_task=True)
                            if task:
                                active_callbacks.add(task)
                                task.add_done_callback(active_callbacks.remove)

            except asyncio.CancelledError:
                logging.info("pulseCounter for di=" + str(di) + " cancelled.")
                break

            except StopAsyncIteration:
                logging.info("pulseCounter for di=" + str(di) + " ended.")
                break

            except Exception as e:
                logging.error("Error receiving pulse for di=" + str(di) + ": " + str(e), exc_info=e)
                # await asyncio.sleep(1)

            ## Loop again
            await asyncio.sleep(1)

        ## Wait for all active callbacks to finish
        while active_callbacks:
            await asyncio.wait(active_callbacks, timeout=1)

    @staticmethod
    def _cast_pins(pins):
        if isinstance(pins, int):
            return [pins]
        elif isinstance(pins, list):
            return pins
        else:
            raise ValueError(f"Invalid type for pins: {type(pins)}. Must be int or list.")
            # logging.error(f"Invalid type for pins: {type(pins)}. Must be int or list."
            # return None

    @staticmethod
    def _cast_values(values):
        if isinstance(values, int):
            return [bool(values)]
        elif isinstance(values, list):
            return [bool(v) for v in values]
        else:
            raise ValueError(f"Invalid type for values: {type(values)}. Must be bool or list.")

    @staticmethod
    def _cast_ao_values(values):
        if isinstance(values, int):
            return [float(values)]
        elif isinstance(values, float):
            return [values]
        elif isinstance(values, list):
            return [float(v) for v in values]
        else:
            raise ValueError(f"Invalid type for values: {type(values)}. Must be float or list.")
        
    def _cast_ao_pin_values(self, pins, values):
        pins = self._cast_pins(pins)
        values = self._cast_ao_values(values)

        if len(pins) != len(values):
            if len(values) == 1:
                values = [values[0]] * len(pins)
            else:
                raise ValueError("Analogue output and value lists are not the same length.")

        return pins, values
    
    def _cast_pin_values(self, pins, values):
        pins = self._cast_pins(pins)
        values = self._cast_values(values)

        if len(pins) != len(values):
            if len(values) == 1:
                values = [values[0]] * len(pins)
            else:
                raise ValueError("Digital output and value lists are not the same length.")

        return pins, values
    
    @cli_command()
    def test_comms(self, message: str = "Comms Check Message") -> Optional[str]:
        """Test connection by sending a basic echo response to platform interface container.

        Parameters
        ----------
        message : str
            Message to send to platform interface to have echo'd as a response

        Returns
        -------
        str
            The response from platform interface.
        """
        return self.make_request("TestComms", platform_iface_pb2.TestCommsRequest(message=message), response_field="response")

    @cli_command()
    @maybe_async()
    def get_di(self, di: int | list[int]) -> Optional[list[bool]]:
        """Get digital input values.

        Parameters
        ----------
        di : int or list
            Pin numbers to get the values of. Can be a single pin number or a list of pin numbers.

        Returns
        -------
        list[bool]
            List of digital input values.
        """
        pins = self._cast_pins(di)
        return self.make_request("getDI", platform_iface_pb2.getDIRequest(di=pins), response_field="di")

    async def get_di_async(self, di: int | list[int]) -> list[bool]:
        pins = self._cast_pins(di)
        return await self.make_request_async("getDI", platform_iface_pb2.getDIRequest(di=pins), response_field="di")

    @cli_command()
    @maybe_async()
    def get_ai(self, ai: int | list[int]) -> Optional[list[float]]:
        """Get analogue input values.

        Parameters
        ----------
        ai : int or list
            Pin numbers to get the values of. Can be a single pin number or a list of pin numbers.

        Returns
        -------
        list[float]
            List of analogue input values.
        """
        # Above section is to facilitate the following:
        # get_ai(1)
        # get_ai([1,4,2])

        # Proposal: get_ai(*pins)
        # allows for get_ai(1, 2, 3) or get_ai(1) or get_ai(*[1, 2, 3])

        pins = self._cast_pins(ai)
        return self.make_request("getAI", platform_iface_pb2.getAIRequest(ai=pins), response_field="ai")

    async def get_ai_async(self, ai: int | list[int]) -> list[float]:
        pins = self._cast_pins(ai)
        return await self.make_request_async("getAI", platform_iface_pb2.getAIRequest(ai=pins), response_field="ai")

    @cli_command()
    @maybe_async()
    def get_do(self, do: int | list[int]) -> Optional[list[bool]]:
        """Get digital output values.

        Parameters
        ----------
        do : int or list
            Pin numbers to get the values of. Can be a single pin number or a list of pin numbers.

        Returns
        -------
        list[bool]
            List of digital output values.
        """
        pins = self._cast_pins(do)
        return self.make_request("getDO", platform_iface_pb2.getDORequest(do=pins), response_field="do")

    async def get_do_async(self, do: int | list[int]) -> list[float]:
        pins = self._cast_pins(do)
        return await self.make_request_async("getDO", platform_iface_pb2.getDORequest(do=pins), response_field="do")

    @cli_command()
    @maybe_async()
    def set_do(self, do: int | list[int], value: int | list[int]) -> Optional[list[bool]]:
        """Set digital output values.

        Parameters
        ----------
        do : int or list
            Pin numbers to set the values of. Can be a single pin number or a list of pin numbers.
        value : int or list
            Values to set the pins to. Can be a single value or a list of values.
            If a single value is provided, all pins will be set to that value.

            .. note::
                The length of the `do` and `value` lists must be the same!

        Returns
        -------
        list[bool]
            ??? Not sure what this is
        """
        pins, values = self._cast_pin_values(do, value)
        return self.make_request("setDO", platform_iface_pb2.setDORequest(do=pins, value=values), response_field="do")

    async def set_do_async(self, do, value):
        pins, values = self._cast_pin_values(do, value)
        return await self.make_request_async("setDO", platform_iface_pb2.setDORequest(do=pins, value=values), response_field="do")

    @cli_command()
    @maybe_async()
    def schedule_do(self, do: int | list[int], value: bool | list[bool], in_secs: int) -> None:
        """Schedule digital output values.

        Parameters
        ----------
        do : int or list
            Pin numbers to set the values of. Can be a single pin number or a list of pin numbers.
        value : int or list
            Values to set the pins to. Can be a single value or a list of values.
            If a single value is provided, all pins will be set to that value.
        in_secs : int
            Time in seconds to schedule the change in digital output values. Must be positive.

        Returns
        -------
        Nothing??
        """
        if not isinstance(in_secs, int) or in_secs < 0:
            raise ValueError(f"Invalid value for in_secs: {in_secs}. Must be a positive integer.")

        pins, values = self._cast_pin_values(do, value)

        # Above section is to facilitate the following:
        # schedule_do(1, 1, 1) => [1],[1],1
        # schedule_do([1,4,2], 0, 1) => [1,4,2], [0,0,0], 1
        # schedule_do([1,4,2], [0,1,0], 1) => [1,4,2], [0,1,0], 1

        return self.make_request("scheduleDO", platform_iface_pb2.scheduleDORequest(do=pins, value=values, time_secs=in_secs), response_field="do")

    async def schedule_do_async(self, do: int | list[int], value: bool | list[bool], in_secs: int) -> None:
        if not isinstance(in_secs, int) or in_secs < 0:
            raise ValueError(f"Invalid value for in_secs: {in_secs}. Must be a positive integer.")

        pins, values = self._cast_pin_values(do, value)
        return await self.make_request_async("scheduleDO", platform_iface_pb2.scheduleDORequest(do=pins, value=values, time_secs=in_secs), response_field="do")

    @cli_command()
    @maybe_async()
    def get_ao(self, ao: int | list[int]) -> Optional[list[float]]:
        """Get analogue output values.

        Parameters
        ----------
        ao : int or list
            Pin numbers to get the values of. Can be a single pin number or a list of pin numbers.

        Returns
        -------
        list[float]
            List of analogue output values.
        """
        pins = self._cast_pins(ao)
        return self.make_request("getAO", platform_iface_pb2.getAORequest(ao=pins), response_field="ao")

    async def get_ao_async(self, ao: int | list[int]) -> list[float]:
        pins = self._cast_pins(ao)
        return await self.make_request_async("getAO", platform_iface_pb2.getAORequest(ao=pins), response_field="ao")

    @cli_command()
    @maybe_async()
    def set_ao(self, ao: int | list[int], value: float | list[float]) -> Optional[list[float]]:
        """Set analogue output values.

        Parameters
        ----------
        ao : int or list[int]
            Pin numbers to set the values of. Can be a single pin number or a list of pin numbers.

        value : bool or list[bool]
            Values to set the pins to. Can be a single value or a list of values.
            If a single value is provided, all pins will be set to that value.

        Returns
        -------
        list[bool]
            List of ouputs set??"""

        # if not isinstance(value, list):
        #     value = [value]
        pins, values = self._cast_ao_pin_values(ao, value)
        return self.make_request("setAO", platform_iface_pb2.setAORequest(ao=pins, value=values), response_field="ao")

    async def set_ao_async(self, ao: int | list[int], value: float | list[float]) -> list[bool]:
        pins, values = self._cast_ao_pin_values(ao, value)
        return await self.make_request_async("setAO", platform_iface_pb2.setAORequest(ao=pins, value=values), response_field="ao")

    @cli_command()
    @maybe_async()
    def schedule_ao(self, ao: int | list[int], value: bool | list[bool], in_secs: int) -> None:
        # fixme: should this function even exist?? setting analog outputs?
        if not isinstance(in_secs, int) or in_secs < 0:
            raise ValueError(f"Invalid value for in_secs: {in_secs}. Must be a positive integer.")

        pins, values = self._cast_ao_pin_values(ao, value)

        # Above section is to facilitate the following:
        # schedule_ao(1, 1, 1) => [1],[1],1
        # schedule_ao([1,4,2], 0, 1) => [1,4,2], [0,0,0], 1
        # schedule_ao([1,4,2], [0,1,0], 1) => [1,4,2], [0,1,0], 1
        return self.make_request(
            "scheduleAO",
            platform_iface_pb2.scheduleAORequest(ao=pins, value=values, time_secs=in_secs),
            response_field="ao"
        )

    async def schedule_ao_async(self, ao: int | list[int], value: bool | list[bool], in_secs: int) -> None:
        if not isinstance(in_secs, int) or in_secs < 0:
            raise ValueError(f"Invalid value for in_secs: {in_secs}. Must be a positive integer.")

        pins, values = self._cast_ao_pin_values(ao, value)
        return await self.make_request_async(
            "scheduleAO",
            platform_iface_pb2.scheduleAORequest(ao=pins, value=values, time_secs=in_secs),
            response_field="ao"
        )

    @cli_command()
    @maybe_async()
    def get_system_voltage(self) -> float:
        return self.make_request("getInputVoltage", platform_iface_pb2.getInputVoltageRequest(), response_field="voltage")

    async def get_system_voltage_async(self) -> float:
        return await self.make_request_async("getInputVoltage", platform_iface_pb2.getInputVoltageRequest(), response_field="voltage")

    @cli_command()
    @maybe_async()
    def get_system_temperature(self):
        return self.make_request("getTemperature", platform_iface_pb2.getTemperatureRequest(), response_field="temperature")

    async def get_system_temperature_async(self):
        return await self.make_request_async("getTemperature", platform_iface_pb2.getTemperatureRequest(), response_field="temperature")

    @cli_command()
    @maybe_async()
    def get_location(self):
        return self.make_request("getLocation", platform_iface_pb2.getLocationRequest(), response_field="location")

    async def get_location_async(self):
        return await self.make_request_async("getLocation", platform_iface_pb2.getLocationRequest(), response_field="location")

    @cli_command()
    @maybe_async()
    def reboot(self):
        return self.make_request("reboot", platform_iface_pb2.rebootRequest())

    async def reboot_async(self):
        # fixme: should these have async varients?
        return await self.make_request_async("reboot", platform_iface_pb2.rebootRequest())

    @cli_command()
    @maybe_async()
    def shutdown(self):
        return self.make_request("shutdown", platform_iface_pb2.shutdownRequest())

    async def shutdown_async(self):
        # fixme: as above
        return await self.make_request_async("shutdown", platform_iface_pb2.shutdownRequest())

    @cli_command()
    @maybe_async()
    def get_immunity_seconds(self):
        return self.make_request("getShutdownImmunity", platform_iface_pb2.getShutdownImmunityRequest(), response_field="immunity_secs")

    async def get_immunity_seconds_async(self):
        return await self.make_request_async("getShutdownImmunity", platform_iface_pb2.getShutdownImmunityRequest(), response_field="immunity_secs")

    @maybe_async()
    def schedule_startup(self, time_secs: int) -> None:
        return self.make_request("scheduleStartup", platform_iface_pb2.scheduleStartupRequest(time_secs=time_secs), response_field="time_secs")

    async def schedule_startup_async(self, time_secs: int) -> None:
        return await self.make_request_async("scheduleStartup", platform_iface_pb2.scheduleStartupRequest(time_secs=time_secs), response_field="time_secs")

    @cli_command()
    @maybe_async()
    def schedule_shutdown(self, time_secs: int) -> None:
        return self.make_request("scheduleShutdown", platform_iface_pb2.scheduleShutdownRequest(time_secs=time_secs), response_field="time_secs")

    async def schedule_shutdown_async(self, time_secs: int) -> None:
        return await self.make_request_async("scheduleShutdown", platform_iface_pb2.scheduleShutdownRequest(time_secs=time_secs), response_field="time_secs")

    @cli_command()
    @maybe_async()
    def get_io_table(self):
        res = self.make_request("getIoTable", platform_iface_pb2.getIoTableRequest(), response_field="io_table")
        # result = json.loads("".join(self.make_request("getIoTable", platform_iface_pb2.getIoTableRequest())))
        if res is None:
            return None
        string = ""
        for i in res:
            string += i
        result = json.loads(string)
        return result
    
    async def get_io_table_async(self):
        res = await self.make_request_async("getIoTable", platform_iface_pb2.getIoTableRequest(), response_field="io_table")
        # result = json.loads("".join(await self.make_request_async("getIoTable", platform_iface_pb2.getIoTableRequest())))
        if res is None:
            return None
        string = ""
        for i in res:
            string += i
        result = json.loads(string)
        return result

    @cli_command()
    @maybe_async()
    def sync_rtc(self):
        return self.make_request("syncRtcTime", platform_iface_pb2.syncRtcTimeRequest())

    async def sync_rtc_async(self):
        return await self.make_request("syncRtcTime", platform_iface_pb2.syncRtcTimeRequest())
    
    @maybe_async()
    def get_events(self, events_from = 0):
        """Get all events.

        Parameters
        ----------
        events_from : None or int
            Starting event id or timestamp (in milliseconds), defaults to all availible.

        Returns
        -------
        list[Event]
            List of events.
        """
        return self.make_request("getEvents", platform_iface_pb2.getEventsRequest(events_from=events_from), response_field="events")

    async def get_events_async(self, events_from = 0):
        return await self.make_request_async("getEvents", platform_iface_pb2.getEventsRequest(events_from=events_from), response_field="events")
    
    @cli_command()
    @maybe_async()
    def get_di_events(self, di_pin, edge, include_system_events = False, events_from = 0):
        """Get di events.

        Parameters
        ----------
        di : int
            Pin number to check events for.
        edge : "rising" or "falling" or "both"
            The edge to listen to evenets on.
        include_system_events : bool = False
            Whether to include system events like for a doovit the cm4 turning on and off or the io board starting up.
        events_from : None or int
            Starting event id or timestamp (in milliseconds), defaults to all availible.

        Returns
        -------
        list[Event]
            List of events for the given digital input pin.
        """
        rising = False
        falling = False
        if edge == "rising":
            rising = True
        elif edge == "falling":
            falling = True
        elif edge == "both":
            rising = True
            falling = True
        return self.make_request("getDIEvents", platform_iface_pb2.getDIEventsRequest(pin=di_pin, rising=rising, falling=falling, include_system_events=include_system_events, events_from=events_from))

    async def get_di_events_async(self, di_pin, edge, include_system_events = False, events_from = 0):
        rising = False
        falling = False
        if edge == "rising":
            rising = True
        elif edge == "falling":
            falling = True
        elif edge == "both":
            rising = True
            falling = True
        return await self.make_request_async("getDIEvents", platform_iface_pb2.getDIEventsRequest(pin=di_pin, rising=rising, falling=falling, include_system_events=include_system_events, events_from=events_from), response_field="events")


platform_iface = PlatformInterface
pulse_counter = PulseCounter



from datetime import datetime

from typing import Union, Optional, Any

from .element import Element
from .misc import Range, Widget


class NotSet:
    pass

class Variable(Element):
    type = "uiVariable"

    def __init__(
        self,
        name: str,
        display_name: str,
        var_type: str,
        curr_val: Any = NotSet,
        precision: int = None,
        ranges: list[Union[Range, dict]] = None,
        earliest_data_time: Optional[datetime] = None,
        **kwargs
    ):
        # kwargs: verbose_str=verbose_str, show_activity=show_activity, form=form, graphic=graphic, layout=layout
        super().__init__(name, display_name, **kwargs)

        self.var_type = var_type
        self._curr_val = curr_val
        self.precision = precision or kwargs.pop("dec_precision", None)
        self.earliest_data_time = earliest_data_time

        self.ranges = []
        if ranges is not None:
            self.add_ranges(*ranges)

    def to_dict(self):
        result = super().to_dict()
        result['type'] = self.type
        result['varType'] = self.var_type

        curr_val = self.current_value
        if curr_val is not None:
            result['currentValue'] = curr_val

        if self.precision is not None:
            result['decPrecision'] = self.precision

        if self.earliest_data_time is not None:
            if isinstance(self.earliest_data_time, datetime):
                result['earliestDataDate'] = int(self.earliest_data_time.timestamp())
            else:
                result['earliestDataDate'] = self.earliest_data_time

        result["ranges"] = [r.to_dict() for r in self.ranges]
        return result

    @property
    def current_value(self):
        if self._curr_val is NotSet:
            return None
        return self._curr_val

    @current_value.setter
    def current_value(self, val):
        self.update(val)

    def update(self, new_value: Any):
        if self.precision is not None and new_value is not None:
            self._curr_val = round(new_value, self.precision)
        else:
            self._curr_val = new_value

    def recv_ui_state_update(self, state: dict[str, Any]) -> None:
        if self._curr_val is NotSet and "currentValue" in state:
            self.current_value = state["currentValue"]

    def add_ranges(self, *range_val: Range):
        for r in range_val:
            # still support legacy dict passing of range values.
            if isinstance(r, Range):
                self.ranges.append(r)
            elif isinstance(r, dict):
                self.ranges.append(Range.from_dict(r))


class NumericVariable(Variable):
    def __init__(
        self,
        name: str,
        display_name: str,
        curr_val: Union[int, float] = None,
        precision: int = None,
        ranges: list[Union[Range, dict]] = None,
        form: Union[Widget, str, None] = None,
        **kwargs
    ):
        super().__init__(
            name, display_name, var_type="float", curr_val=curr_val,
            precision=precision, ranges=ranges, **kwargs
        )
        self.form = form

    def to_dict(self):
        result = super().to_dict()
        if self.form is not None:
            result["form"] = self.form
        return result

class TextVariable(Variable):
    def __init__(self, name: str, display_name: str, curr_val: str = None, **kwargs):
        # fixme: double check this type
        super().__init__(name, display_name, var_type="string", curr_val=curr_val, **kwargs)


class BooleanVariable(Variable):
    def __init__(self, name: str, display_name: str, curr_val: bool = None, **kwargs):
        super().__init__(name, display_name, var_type="bool", curr_val=curr_val, **kwargs)


class DateTimeVariable(Variable):
    def __init__(self, name: str, display_name: str, curr_val: Union[datetime, int] = None, **kwargs):
        # fixme: double check this type, and how to handle different date / time / datetime
        super().__init__(name, display_name, var_type="time", curr_val=curr_val, **kwargs)


doover_ui_variable = Variable



import inspect

from functools import wraps


def command(name: str = None, description: str = None, setup_api: bool = False):
    def wrapper(func):
        func._is_command = True
        func._command_name = name or func.__name__
        if description:
            func._command_help = description
        elif inspect.getdoc(func):
            func._command_help = inspect.cleandoc(inspect.getdoc(func))
        else:
            func._command_help = ""

        func._command_setup_api = setup_api
        if not hasattr(func, "_command_arg_docs"):
            func._command_arg_docs = dict()

        @wraps(func)
        def inner(*args, **kwargs):
            if setup_api:
                # first arg is self
                args[0].setup_api()
            return func(*args, **kwargs)
        return inner

    return wrapper


def annotate_arg(arg_name: str, description: str):
    def wrapper(func):
        if not hasattr(func, "_command_arg_docs"):
            func._command_arg_docs = dict()

        func._command_arg_docs[arg_name] = description

        @wraps(func)
        def inner(*args, **kwargs):
            return func(*args, **kwargs)
        return inner
    return wrapper


def ignore_alias(func):
    func._is_command = False
    return func


from typing import Union, Any


class Colour:
    blue = "blue"
    yellow = "yellow"
    red = "red"
    green = "green"
    magenta = "magenta"
    limegreen = "limegreen"
    tomato = "tomato"
    orange = "orange"
    purple = "purple"
    grey = "grey"

    @classmethod
    def from_hex(cls, hex_string):
        return hex_string

    @classmethod
    def from_string(cls, value):
        return value  # fixme: this hackiness


class Range:
    def __init__(
        self, label: str = None, min_val: Union[int, float] = None,
        max_val: Union[int, float] = None, colour: "Colour" = Colour.blue, show_on_graph: bool = True
    ):
        self.label = label
        self.min = min_val
        self.max = max_val
        self.colour = colour
        self.show_on_graph = show_on_graph

    def to_dict(self):
        to_return = {
            "min": self.min,
            "max": self.max,
            "colour": self.colour,
            "show_on_graph": self.show_on_graph,
        }
        if self.label:
            to_return["label"] = self.label
        return to_return

    @classmethod
    def from_dict(cls, data: dict[str, Any]):
        return cls(data.get("label"), data["min"], data["max"], Colour.from_string(data["colour"]), data["show_on_graph"])


class Option:
    def __init__(self, name: str, display_name: str):
        self.name = name
        self.display_name = display_name

    def to_dict(self):
        return {
            "name": self.name,
            "displayString": self.display_name,
            "type": "uiElement",
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]):
        return cls(data["name"], data["display_str"])


class Widget:
    linear = "linearGauge"
    radial = "radialGauge"

    @classmethod
    def from_string(cls, value):
        return value  # fixme: this hackiness


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: grpc_stubs/camera_iface.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dgrpc_stubs/camera_iface.proto\x12\x0c\x63\x61mera_iface\"H\n\x0eStreamSettings\x12\x10\n\x03\x66ps\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x12\n\x05scale\x18\x02 \x01(\tH\x01\x88\x01\x01\x42\x06\n\x04_fpsB\x08\n\x06_scale\"q\n\x0eResponseHeader\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x1a\n\rresponse_code\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x14\n\x07message\x18\x03 \x01(\tH\x01\x88\x01\x01\x42\x10\n\x0e_response_codeB\n\n\x08_message\"#\n\x10TestCommsRequest\x12\x0f\n\x07message\x18\x01 \x01(\t\"\\\n\x11TestCommsResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.camera_iface.ResponseHeader\x12\x10\n\x08response\x18\x02 \x01(\t\"\xad\x02\n\x0fSnapshotRequest\x12\x10\n\x08rtsp_uri\x18\x01 \x01(\t\x12\x1a\n\rsnapshot_type\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x1c\n\x0fsnapshot_length\x18\x03 \x01(\x05H\x01\x88\x01\x01\x12\x1b\n\x0eoutput_channel\x18\x04 \x01(\tH\x02\x88\x01\x01\x12\x18\n\x0b\x63\x61mera_name\x18\x05 \x01(\tH\x03\x88\x01\x01\x12:\n\x0fstream_settings\x18\x06 \x01(\x0b\x32\x1c.camera_iface.StreamSettingsH\x04\x88\x01\x01\x42\x10\n\x0e_snapshot_typeB\x12\n\x10_snapshot_lengthB\x11\n\x0f_output_channelB\x0e\n\x0c_camera_nameB\x12\n\x10_stream_settings\"m\n\x10SnapshotResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.camera_iface.ResponseHeader\x12\x15\n\x08response\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\x0b\n\t_response\"\x1a\n\x18IsSnapshotRunningRequest\"z\n\x19IsSnapshotRunningResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.camera_iface.ResponseHeader\x12\x17\n\nis_running\x18\x02 \x01(\x08H\x00\x88\x01\x01\x42\r\n\x0b_is_running2\x95\x02\n\x0b\x63\x61meraIface\x12N\n\tTestComms\x12\x1e.camera_iface.TestCommsRequest\x1a\x1f.camera_iface.TestCommsResponse\"\x00\x12N\n\x0bRunSnapshot\x12\x1d.camera_iface.SnapshotRequest\x1a\x1e.camera_iface.SnapshotResponse\"\x00\x12\x66\n\x11IsSnapshotRunning\x12&.camera_iface.IsSnapshotRunningRequest\x1a\'.camera_iface.IsSnapshotRunningResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'grpc_stubs.camera_iface_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_STREAMSETTINGS']._serialized_start=47
  _globals['_STREAMSETTINGS']._serialized_end=119
  _globals['_RESPONSEHEADER']._serialized_start=121
  _globals['_RESPONSEHEADER']._serialized_end=234
  _globals['_TESTCOMMSREQUEST']._serialized_start=236
  _globals['_TESTCOMMSREQUEST']._serialized_end=271
  _globals['_TESTCOMMSRESPONSE']._serialized_start=273
  _globals['_TESTCOMMSRESPONSE']._serialized_end=365
  _globals['_SNAPSHOTREQUEST']._serialized_start=368
  _globals['_SNAPSHOTREQUEST']._serialized_end=669
  _globals['_SNAPSHOTRESPONSE']._serialized_start=671
  _globals['_SNAPSHOTRESPONSE']._serialized_end=780
  _globals['_ISSNAPSHOTRUNNINGREQUEST']._serialized_start=782
  _globals['_ISSNAPSHOTRUNNINGREQUEST']._serialized_end=808
  _globals['_ISSNAPSHOTRUNNINGRESPONSE']._serialized_start=810
  _globals['_ISSNAPSHOTRUNNINGRESPONSE']._serialized_end=932
  _globals['_CAMERAIFACE']._serialized_start=935
  _globals['_CAMERAIFACE']._serialized_end=1212
# @@protoc_insertion_point(module_scope)
